/* -*-c-*- */
/* Copyright Olivier Chapuis (CNRS), Matthieu Langet (CNRS), 2005.
 *
 * <chapuis@lri.fr> <mathieu.langet@lri.fr>
 *
 * This software is a computer program whose purpose is to log the
 * window management activity into some files.
 *
 * This software is governed by the CeCILL  license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 *
 */


#include "config.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>

#include "atoms.h"

/* checking atoms */
atom_t atom_net_check =     {None, XA_WINDOW};
atom_t atom_net_supported = {None, XA_ATOM};
atom_t atom_win_check =     {None, XA_CARDINAL};
atom_t atom_win_supported = {None, XA_ATOM};

/* "root" atoms */
atom_t atom_client_list =              {None, None};
atom_t atom_net_client_list =          {None, XA_WINDOW};
atom_t atom_net_client_list_stacking = {None, XA_WINDOW};
atom_t atom_win_client_list =          {None, XA_CARDINAL};
atom_t atom_net_active_window =        {None, XA_WINDOW};
atom_t atom_net_current_desktop =      {None, XA_CARDINAL};
atom_t atom_net_number_of_desktops =   {None, XA_CARDINAL};
atom_t atom_net_desktop_viewport =     {None, XA_CARDINAL};
atom_t atom_net_desktop_geometry =     {None, XA_CARDINAL};

atom_t atom_net_desktop_names =    {None, None};
atom_t atom_net_desktop_layout =   {None, XA_CARDINAL};
atom_t atom_net_showing_desktop =  {None, XA_CARDINAL};
atom_t atom_net_workarea =         {None, XA_CARDINAL};
atom_t atom_net_virtual_roots =    {None, XA_WINDOW};

/* "window" atoms */
atom_t atom_wm_desktop =      {None, XA_CARDINAL};
atom_t atom_wm_state =        {None, XA_CARDINAL};
atom_t atom_net_wm_state =    {None, XA_ATOM};
atom_t atom_win_state =       {None, XA_CARDINAL};
atom_t atom_win_hints =       {None, XA_CARDINAL};
atom_t atom_win_layer =       {None, XA_CARDINAL};

atom_t atom_net_wm_allowed_actions = {None, XA_ATOM};
atom_t atom_net_wm_strut =          {None, XA_CARDINAL};
atom_t atom_net_wm_partial_strut =  {None, XA_CARDINAL};
atom_t atom_net_wm_icon_geometry =  {None, XA_CARDINAL};
atom_t atom_net_wm_icon =           {None, XA_CARDINAL};
atom_t atom_net_wm_handle_icons =   {None, XA_CARDINAL};
atom_t atom_net_wm_user_time =      {None, XA_CARDINAL};
atom_t atom_net_wm_extents =        {None, XA_CARDINAL};
atom_t atom_kde_net_frame_strut =   {None, XA_CARDINAL};

atom_t atom_net_wm_window_type = {None, XA_ATOM};
atom_t atom_win_window_type = {None, XA_ATOM};
atom_t atom_net_wm_pid =         {None, XA_CARDINAL};
atom_t atom_mmpl_pid =        {None, XA_CARDINAL};
atom_t atom_mmpl_parent_pid = {None, XA_CARDINAL};
atom_t atom_client_machine =  {None, XA_STRING};

atom_t atom_wm_name              = {XA_WM_NAME, XA_STRING};
atom_t atom_net_wm_name          = {None, None};
atom_t atom_net_wm_visible_name  = {None, None};
atom_t atom_wm_icon_name         = {XA_WM_ICON_NAME, XA_STRING};
atom_t atom_net_wm_icon_name          = {None, None};
atom_t atom_net_wm_visible_icon_name  = {None, None};

/* _net_wm_state state */
atom_t atom_net_wm_state_below = {None, XA_ATOM};
atom_t atom_net_wm_state_above = {None, XA_ATOM};
atom_t atom_net_wm_state_stays_on_top = {None, XA_ATOM};
atom_t atom_net_wm_state_fullscreen = {None, XA_ATOM};
atom_t atom_net_wm_state_hidden = {None, XA_ATOM};
atom_t atom_net_wm_state_horz = {None, XA_ATOM};
atom_t atom_net_wm_state_ = {None, XA_ATOM};
atom_t atom_net_wm_state_maximized_horiz = {None, XA_ATOM};
atom_t atom_net_wm_state_maximized_horz = {None, XA_ATOM};
atom_t atom_net_wm_state_maximized_vert = {None, XA_ATOM};
atom_t atom_net_wm_state_modal = {None, XA_ATOM};
atom_t atom_net_wm_state_shaded = {None, XA_ATOM};
atom_t atom_net_wm_state_skip_pager = {None, XA_ATOM};
atom_t atom_net_wm_state_skip_taskbar = {None, XA_ATOM};
atom_t atom_net_wm_state_sticky = {None, XA_ATOM};

/* allowed action */
atom_t atom_net_wm_action_move = {None, XA_ATOM};
atom_t atom_net_wm_action_resize = {None, XA_ATOM};
atom_t atom_net_wm_action_minimize = {None, XA_ATOM};
atom_t atom_net_wm_action_shade = {None, XA_ATOM};
atom_t atom_net_wm_action_stick = {None, XA_ATOM};
atom_t atom_net_wm_action_maximize_horz = {None, XA_ATOM};
atom_t atom_net_wm_action_maximize_vert = {None, XA_ATOM};
atom_t atom_net_wm_action_fullscreen = {None, XA_ATOM};
atom_t atom_net_wm_action_change_desktop = {None, XA_ATOM};
atom_t atom_net_wm_action_close = {None, XA_ATOM};

#define XIA(a) XInternAtom(disp, a, False);

void atom_init0(Display *disp)
{
	Atom utf8;

	atom_net_check.atom = XIA("_NET_SUPPORTING_WM_CHECK");
	atom_net_supported.atom = XIA("_NET_SUPPORTED");
	atom_win_check.atom = XIA("_WIN_SUPPORTING_WM_CHECK");
	atom_win_supported.atom = XIA("_WIN_PROTOCOLS");

	/* should init these atoms here */
	utf8 = XIA("UTF8_STRING");
	atom_net_wm_name.type = utf8;
	atom_net_wm_visible_name.type = utf8;
	atom_net_wm_icon_name.type = utf8;
	atom_net_wm_visible_icon_name.type = utf8;
	atom_net_wm_name.atom = XIA("_NET_WM_NAME");
	atom_net_wm_visible_name.atom =  XIA("_NET_WM_VISIBLE_NAME");
	atom_net_wm_icon_name.atom = XIA("_NET_WM_ICON_NAME");
	atom_net_wm_visible_icon_name.atom =  XIA("_NET_WM_VISIBLE_ICON_NAME");
}

void atom_init(Display *disp, Bool ewmh_compliant)
{

	atom_net_client_list.atom = XIA("_NET_CLIENT_LIST");
	atom_net_client_list_stacking.atom = XIA("_NET_CLIENT_LIST_STACKING");
	atom_win_client_list.atom = XIA("_WIN_CLIENT_LIST");

	if (ewmh_compliant)
	{
		atom_client_list = atom_net_client_list;
		atom_net_current_desktop.atom = XIA("_NET_CURRENT_DESKTOP");
		atom_wm_desktop.atom = XIA("_NET_WM_DESKTOP");
	}
	else
	{
		atom_client_list = atom_win_client_list;
		atom_net_current_desktop.atom = atom_wm_desktop.atom =
			XIA("_WIN_WORKSPACE");
	}

	atom_net_active_window.atom = XIA("_NET_ACTIVE_WINDOW");
	atom_net_number_of_desktops.atom = XIA("_NET_NUMBER_OF_DESKTOPS");
	atom_net_desktop_viewport.atom = XIA("_NET_DESKTOP_VIEWPORT");
	atom_net_desktop_geometry.atom = XIA("_NET_DESKTOP_GEOMETRY");
	
	atom_net_desktop_names.atom = XIA("_NET_DESKTOP_NAMES");
	atom_net_desktop_layout.atom = XIA("_NET_DESKTOP_LAYOUT");
	atom_net_showing_desktop.atom = XIA("_NET_SHOWING_DESKTOP");
	atom_net_workarea.atom = XIA("_NET_WORKAREA");
	atom_net_virtual_roots.atom = XIA("_NET_VIRTUAL_ROOTS");

	atom_wm_state.atom = XIA("WM_STATE");
	atom_wm_state.type = atom_wm_state.atom;
	atom_net_wm_state.atom = XIA("_NET_WM_STATE");
	atom_win_state.atom = XIA("_WIN_STATE");
	atom_win_hints.atom = XIA("_WIN_HINTS");
	atom_win_layer.atom = XIA("_WIN_LAYER");

	atom_net_wm_allowed_actions.atom = XIA("_NET_WM_ALLOWED_ACTIONS");
	atom_net_wm_strut.atom = XIA("_NET_WM_STRUT");
	atom_net_wm_partial_strut.atom = XIA("_NET_WM_PARTIAL_STRUT");
	atom_net_wm_icon_geometry.atom = XIA("_NET_WM_ICON_GEOMETRY");
	atom_net_wm_icon.atom = XIA("_NET_WM_ICON");
	atom_net_wm_handle_icons.atom = XIA("_NET_WM_HANDLE_ICONS");
	atom_net_wm_user_time.atom = XIA("_NET_WM_USER_TIME");
	atom_net_wm_extents.atom = XIA("_NET_FRAME_EXTENTS");
	atom_kde_net_frame_strut.atom = XIA("_KDE_NET_WM_FRAME_STRUT");

	atom_net_wm_window_type.atom = XIA("_NET_WM_WINDOW_TYPE");
	atom_win_window_type.atom = XIA("_WIN_WINDOW_TYPE");

	atom_net_wm_pid.atom =  XIA("_NET_WM_PID");
	atom_mmpl_pid.atom = XIA("_MMPL_PID");
	atom_mmpl_parent_pid.atom = XIA("_MMPL_PARENT_PID");

	atom_client_machine.atom = XIA("WM_CLIENT_MACHINE");

	atom_net_wm_state_below.atom = XIA("_NET_WM_STATE_BELOW");
	atom_net_wm_state_above.atom = XIA("_NET_WM_STATE_ABOVE");
	atom_net_wm_state_stays_on_top.atom = XIA("_NET_WM_STATE_STAYS_ON_TOP");
	atom_net_wm_state_fullscreen.atom = XIA("_NET_WM_STATE_FULLSCREEN");
	atom_net_wm_state_hidden.atom = XIA("_NET_WM_STATE_HIDDEN");
	atom_net_wm_state_maximized_horz.atom = XIA(
		"_NET_WM_STATE_MAXIMIZED_HORZ");
	atom_net_wm_state_maximized_horiz.atom = XIA(
		"_NET_WM_STATE_MAXIMIZED_HORIZ");
	atom_net_wm_state_maximized_vert.atom = XIA(
		"_NET_WM_STATE_MAXIMIZED_VERT");
	atom_net_wm_state_modal.atom = XIA("_NET_WM_STATE_MODAL");
	atom_net_wm_state_shaded.atom = XIA("_NET_WM_STATE_SHADED");
	atom_net_wm_state_skip_pager.atom = XIA("_NET_WM_STATE_SKIP_PAGER");
	atom_net_wm_state_skip_taskbar.atom = XIA("_NET_WM_STATE_SKIP_TASKBAR");
	atom_net_wm_state_sticky.atom = XIA("_NET_WM_STATE_STICKY");

	atom_net_wm_action_move.atom = XIA("NET_WM_ACTION_MOVE");
	atom_net_wm_action_resize.atom = XIA("_NET_WM_ACTION_RESIZE");
	atom_net_wm_action_minimize.atom = XIA("_NET_WM_ACTION_MINIMIZE");
	atom_net_wm_action_shade.atom = XIA("_NET_WM_ACTION_SHADE");
	atom_net_wm_action_stick.atom = XIA("_NET_WM_ACTION_STICK");
	atom_net_wm_action_maximize_horz.atom =
		XIA("_NET_WM_ACTION_MAXIMIZE_HORZ");
	atom_net_wm_action_maximize_vert.atom =
		XIA("_NET_WM_ACTION_MAXIMIZE_VERT");
	atom_net_wm_action_fullscreen.atom = XIA("_NET_WM_ACTION_FULLSCREEN");
	atom_net_wm_action_change_desktop.atom =
		XIA("_NET_WM_ACTION_CHANGE_DESKTOP");
	atom_net_wm_action_close.atom = XIA("_NET_WM_ACTION_CLOSE");
}
