/* -*-c-*- */
/* Copyright Olivier Chapuis (CNRS), Matthieu Langet (CNRS), 2005.
 *
 * <chapuis@lri.fr> <mathieu.langet@lri.fr>
 *
 * This software is a computer program whose purpose is to log the
 * window management activity into some files.
 *
 * This software is governed by the CeCILL  license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 *
 */

#include "config.h"

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <time.h>
#include <signal.h>
#include <errno.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/Xproto.h>
#include <X11/Xlocale.h>

#include "FRecord.h"

#include "global.h"
#include "atoms.h"
#include "xget.h"
#include "xwindow.h"

#if IS_WMWATCH
#include "watch.h"
#elif IS_WMTRACE
#include "trace.h"
#endif

Display *disp;
atom_t atom_selection;

/* options */
Bool Development = False;
Bool trackMouseMotion = True;
Bool flushFile = False;
int privacyLevel = 1;

/* program name for error msg */
char *myName;

int last_event_event_time = CurrentTime;
int last_record_event_time = CurrentTime;

extern xwindow list_xwindow[1024];
extern int list_xwindow_size;
extern unsigned long xinput_mask;

static FRecordRecordDescription *frd = NULL;
static Bool get_xioerror = False;

void exit_wmwatch()
{
	if (!get_xioerror && frd != NULL)
	{
		FRecordStopRecording(frd);
	}
	if (!get_xioerror && frd != NULL)
	{
		XCloseDisplay(disp);
	}

	close_sender();

	exit(0);
}

int myErrorHandler(Display *dpy, XErrorEvent *e)
{
	char msg[256];

	/* some errors are acceptable BadAtom? */

	if (e->error_code == BadWindow ||
	    e->request_code == X_GetGeometry ||
	    e->error_code == BadDrawable)
	{
		return 0;
	}

	send_xerror(e);
	XGetErrorText(dpy, e->error_code, msg, sizeof(msg));
	fprintf(stderr, "%s XError: %d (%s)\n", myName, e->error_code, msg);
	fprintf(stderr, "   Major opcode of failed request:  %d\n",
		e->request_code);
	fprintf(stderr, "   Minor opcode of failed request:  %d \n",
		e->minor_code);
	fprintf(stderr, "   Resource id of failed request:  0x%lx \n",
		e->resourceid);

	if (Development)
	{
		fprintf(stderr, " Leaving a core dump now\n");
		abort();
		/* exit if this fails */
		exit(99);
	}

	return 0 ;
}

int xioerror(Display *disp)
{
	/* killed by XKillClient or Serveur shutdown */
	fprintf(stderr, "%s: killed by XKillClient or Serveur shutdown\n",
		myName);
	get_xioerror = True;

	exit_wmwatch();
	exit(0);

	return 0;
}

/*
 * Imprime la liste des fenetres avec les infos suivantes  :
 * adresse de la fenetre, numero de bureau, pid, x, y, w, h, nom de la machine, nom de la fenetre, type de la fenetre
 */
void print_list_xwindow() {
  xwindow xwin;
  int i;
  printf("--------------------------------------------------\n");
  for (i = 0; i < list_xwindow_size; i++) {
    xwin = list_xwindow[i];
    printf("%lx %lu %ld %d %d %d %d %s %s %s\n",xwin.window_,xwin.desktop_,xwin.pid_,xwin.x_,xwin.y_,xwin.w_, xwin.h_,(xwin.machine_)?xwin.machine_:"None",xwin.name_, xwin.type_);
  }
}


static
void add_the_root_window()
{
	Window root = DefaultRootWindow(disp);
	int id;
	XClassHint classhints;

	classhints.res_name = NULL;
	classhints.res_class = NULL;
	id = create_xwindow(
		root, root, None,
		0xFFFFFFFE, /* desktop */
		0, /* pid, get the X server pid? */
		0,0, /* x,y */
		DisplayWidth (disp, DefaultScreen(disp)),
		DisplayHeight(disp, DefaultScreen(disp)),
		NULL,
		NULL, /* machine gethostname()? */
		NULL, classhints,
		"Root Window", 0,
		"_NET_ROOT_WINDOW", /* type */
		0, /* state */
		0, /*layer */
		NULL, /*strut*/
		NULL, /*partial_strut*/
		0,0,0,0, /*icon geometry*/
		0, /*handle_icon*/
		-1, /*allowed_action*/
		MM_ROOTWIN, CurrentTime);

	if (id != ROOT_ID || ROOT_ID != 0)
	{
		fprintf(stderr,"Internal Error\n");
		exit(36);
	}
}

/*
 *
 * XRecord stuff
 *
 */

void
record_send_event(FRecordInterceptData *xrecintd)
{
	FRecordDatum *xrec_data  = (FRecordDatum *) (xrecintd->data) ;
	int event_type = xrec_data->type ;

	/* xrec_data->event.u.* is described in Xproto.h */
	switch(event_type)
	{
	case ButtonPress:
	case ButtonRelease: 
	case MotionNotify:
	case KeyPress:
	case KeyRelease:
		last_record_event_time =
			xrec_data->event.u.keyButtonPointer.time;
		break;
	case SelectionRequest:
		last_record_event_time =
			xrec_data->event.u.selectionRequest.time;
		break;
	default:
		break;
	}
  
	switch(event_type)
	{
	case ButtonPress:
		set_button_event(
			xrec_data->event.u.keyButtonPointer.time,
			xrec_data->event.u.keyButtonPointer.event,
			wmButtonPress, xrec_data->event.u.keyButtonPointer.child,
			xrec_data->event.u.keyButtonPointer.rootX,
			xrec_data->event.u.keyButtonPointer.rootY,
			xrec_data->event.u.u.detail,
			xrec_data->event.u.keyButtonPointer.state);
		break;
	case ButtonRelease:
		set_button_event(
			xrec_data->event.u.keyButtonPointer.time,
			xrec_data->event.u.keyButtonPointer.event,
			wmButtonRelease,
			xrec_data->event.u.keyButtonPointer.child,
			xrec_data->event.u.keyButtonPointer.rootX,
			xrec_data->event.u.keyButtonPointer.rootY,
			xrec_data->event.u.u.detail,
			xrec_data->event.u.keyButtonPointer.state);
		break;
	case MotionNotify:
		set_button_event(
			xrec_data->event.u.keyButtonPointer.time,
			xrec_data->event.u.keyButtonPointer.event,
			wmButtonMotion,
			xrec_data->event.u.keyButtonPointer.child,
			xrec_data->event.u.keyButtonPointer.rootX,
			xrec_data->event.u.keyButtonPointer.rootY,
			xrec_data->event.u.u.detail,
			xrec_data->event.u.keyButtonPointer.state);
		break;
	case KeyPress:
	{
		KeySym key_sym = XKeycodeToKeysym(
			disp, xrec_data->event.u.u.detail, 0);
		set_key_event(
			xrec_data->event.u.keyButtonPointer.time, wmKeyPress,
			xrec_data->event.u.keyButtonPointer.event,
			key_sym,
			xrec_data->event.u.keyButtonPointer.state);
		break;
	}
	case KeyRelease:
	{
		KeySym key_sym = XKeycodeToKeysym(
			disp, xrec_data->event.u.u.detail, 0);
		set_key_event(
			xrec_data->event.u.keyButtonPointer.time, wmKeyRelease,
			xrec_data->event.u.keyButtonPointer.event,
			key_sym,
			xrec_data->event.u.keyButtonPointer.state);
		break;
	}
	case SelectionRequest:
		if (xrec_data->event.u.selectionRequest.property !=
		    atom_selection.atom)
		{
			set_selection_request(
				xrec_data->event.u.selectionRequest.time, 
				xrec_data->event.u.selectionRequest.requestor,
				xrec_data->event.u.selectionRequest.owner);
		}
		break;
	default:
		break;
	}
}


void
record_dispatch(XPointer xpointer_data, FRecordInterceptData *data) {
  FRecordDatum *xrec_data;
  int data_type = 0;

  if ( (!data->data) || (data==NULL) ) {
    FRecordFreeData(data);
    return;
  }

  xrec_data = (FRecordDatum *) (data->data) ;

  if(data->data_len) {
    data_type = xrec_data->type;
  }

  switch(data->category) {
    case FRecordFromServer:
      if(data_type >  X_Reply ) {
        record_send_event(data) ;
      }
      break;
    default:
      break;
  }
  FRecordFreeData(data);
}

static
void process_event()
{
	XEvent e;
	int id;
	Window root = DefaultRootWindow(disp);

	XNextEvent(disp, &e);

	switch (e.type)
	{
#if IS_WMTRACE
	/* for override redirect windows and maybe others special
	 * window */
	case MapNotify:
		last_event_event_time = CurrentTime;
		checkadd_unmanaged_xwindow(
			CurrentTime, e.xmap.window, e.xmap.override_redirect);
		break;	
	case UnmapNotify:
		last_event_event_time = CurrentTime;
		checkremove_unmanaged_xwindow(CurrentTime, e.xunmap.window);
		break;
	case MappingNotify:
		last_event_event_time = CurrentTime;
		update_shortcut_modifiers(disp);
		break;
#endif
	case MotionNotify:
		last_event_event_time = e.xmotion.time;;
		// handled by XRecord
		break;
	case KeyPress:
	case KeyRelease:
		last_event_event_time = e.xkey.time;
		break;
	case ButtonPress:
	case ButtonRelease:
		last_event_event_time = e.xbutton.time;
		break;
	case EnterNotify:
	case LeaveNotify:
		last_event_event_time = e.xcrossing.time;
		break;
	case FocusIn:
		if ((id = get_id_xwindow(e.xfocus.window)) != -1) {
			set_focusin_xwindow(CurrentTime, id);
		}
		break;
	case FocusOut:
		if ((id = get_id_xwindow(e.xfocus.window)) != -1) {
			set_focusout_xwindow(CurrentTime, id);
		}
		break;
	case ConfigureNotify:
	{
		while(XCheckTypedWindowEvent(
			      disp, e.xconfigure.event, ConfigureNotify,
			      &e))
		{
			/* flush */
		}
		id = get_id_xwindow(e.xconfigure.window);
		if (id == -1)
		{
			id = get_id_xwindow_by_frame(e.xconfigure.window);
			if (id != -1 &&
			    list_xwindow[id].frame_ != list_xwindow[id].window_)
			{
				while(XCheckTypedWindowEvent(
					      disp, list_xwindow[id].window_,
					      ConfigureNotify, &e))
				{
					/* flush */
				}
			}
		}
		else if (list_xwindow[id].frame_ != None &&
			 list_xwindow[id].frame_ != list_xwindow[id].window_)
		{
			while(XCheckTypedWindowEvent(
					      disp, list_xwindow[id].frame_,
					      ConfigureNotify, &e))
			{
				/* flush */
			}
		}
		if (id != -1)
		{
			set_geometry_xwindow(CurrentTime,id);
		}
		else if ((id = get_id_xwindow(e.xconfigure.event)) != -1 &&
			 id != ROOT_ID) {
			fprintf(stderr, "CURIOUS ConfigureNotify\n");
			set_geometry_xwindow(CurrentTime,id);
		}
		break;
	}
	case SelectionClear:
		last_event_event_time = e.xselectionclear.time;
		break;
	case SelectionRequest:
		last_event_event_time = e.xselectionrequest.time;
		break;
	case SelectionNotify:
	{
		atom_t atom;

		last_event_event_time = e.xselection.time;
		atom.atom = e.xselection.property;
		atom.type = e.xselection.target;
		if (atom.atom != None && atom.type != None)
		{
			set_selection_notify(
				e.xselection.time, e.xselection.requestor,
				atom);
		}
		break;
	}
	case PropertyNotify:
		last_event_event_time = e.xproperty.time;
		if (e.xproperty.window == root &&
		    e.xproperty.atom == atom_client_list.atom)
		{
			check_xwindow(e.xproperty.time);
		}
		else if (e.xproperty.window == root &&
			 e.xproperty.atom == atom_net_current_desktop.atom)
		{
			set_current_desktop(e.xproperty.time);
		}
		else if (e.xproperty.atom == atom_net_active_window.atom &&
			 e.xproperty.window == root)
		{
			set_active_xwindow(e.xproperty.time);
		}
		else if (e.xproperty.window == root &&
			 e.xproperty.atom == atom_net_client_list_stacking.atom)
		{
			set_stacking_list(e.xproperty.time);
		}
		else if (e.xproperty.window == root &&
			 e.xproperty.atom == atom_net_number_of_desktops.atom)
		{
			set_number_of_desktops(e.xproperty.time);
		}
		else if (e.xproperty.window == root &&
			 e.xproperty.atom == atom_net_desktop_viewport.atom)
		{
			set_desktop_viewport(e.xproperty.time);
		}
		else if (e.xproperty.window == root &&
			 e.xproperty.atom == atom_net_desktop_geometry.atom)
		{
			set_desktop_geometry(e.xproperty.time);
		}
		else if (e.xproperty.window == root &&
			 e.xproperty.atom == atom_net_desktop_names.atom)
		{
			set_desktop_names(e.xproperty.time);
		}
		else if (e.xproperty.window == root &&
			 e.xproperty.atom == atom_net_desktop_layout.atom)
		{
			set_desktop_layout(e.xproperty.time);
		}
		else if (e.xproperty.window == root &&
			 e.xproperty.atom == atom_net_workarea.atom)
		{
			set_workarea(e.xproperty.time);
		}
		else if (e.xproperty.window == root &&
			 e.xproperty.atom == atom_net_virtual_roots.atom)
		{
			set_virtual_roots(e.xproperty.time);
		}
		else if (e.xproperty.window == root &&
			 e.xproperty.atom == atom_net_showing_desktop.atom)
		{
			set_showing_desktop(e.xproperty.time);
		}

		/* now we need a window to operate */
		if (e.xproperty.window == root ||
		    (id = get_id_xwindow(e.xproperty.window)) == -1)
		{
			break;
		}

		if (e.xproperty.atom == atom_wm_desktop.atom)
		{
			set_desktop_xwindow(e.xproperty.time, id);
		}
		else if (e.xproperty.atom == atom_net_wm_state.atom ||
			 e.xproperty.atom == atom_win_state.atom)
		{
			set_state_xwindow(e.xproperty.time,id);
		}
		else if (e.xproperty.atom == atom_win_layer.atom)
		{
			set_layer_xwindow(e.xproperty.time,id);
		}
		else if (e.xproperty.atom == atom_net_wm_visible_name.atom ||
			 (list_xwindow[id].is_net_name_ != 2 &&
			  e.xproperty.atom == atom_net_wm_name.atom) ||
			 (list_xwindow[id].is_net_name_ == 0 &&
			  e.xproperty.atom == atom_wm_name.atom))
		{
			set_name_xwindow(e.xproperty.time, id);
		}
		else if (e.xproperty.atom == atom_wm_state.atom)
		{
			set_wm_state_xwindow(e.xproperty.time, id);
		}
		else if (e.xproperty.atom == atom_net_wm_strut.atom)
		{
			set_strut(e.xproperty.time, id);
		}
		else if (e.xproperty.atom == atom_net_wm_partial_strut.atom)
		{
			set_partial_strut(e.xproperty.time, id);
		}
		else if (e.xproperty.atom == atom_net_wm_icon_geometry.atom)
		{
			set_icon_geometry(e.xproperty.time, id);
		}
		else if (e.xproperty.atom == atom_net_wm_handle_icons.atom)
		{
			set_handle_icons(e.xproperty.time, id);
		}
		else if (e.xproperty.atom == atom_net_wm_allowed_actions.atom)
		{
			set_allowed_actions(e.xproperty.time, id);
		}
		break;
	}	
}

static
void usage(Bool verbose)
{
	fprintf(stderr, "usage: %s", myName);
	fprintf(stderr,
		" [-d display]"
		" [-p level]"
		" [-m] [-u] [-h]\n");

	if (!verbose)
	{
		return;
	}

	fprintf(stderr,
		" -d display:   run on <display>\n"
		" -f:           flush file\n"
		" -h:           print this help message\n"
		" -p level:     privacy level 0 or 1\n"
		" -m            do not track mouse motion\n"
		" -u:           for development\n");
}

int main(int argc, char * argv[])
{
  struct sigaction  sigact;
  Bool ewmh_compliant = False;
  Bool win_compliant = False;
  Bool do_exit = False;
  Window win_select;
  int x_fd;
  int fd_width;
  int xgso_count =  0;
  int i;
  char *display_name = NULL;

  
  if (setlocale(LC_CTYPE, "") == NULL)
  {
	  fprintf(stderr,
		  "[%s] ERROR -- Cannot set locale. Please check"
		  " your $LC_CTYPE or $LANG.\n", argv[0]);
  }

  for (i = 1; i < argc; i++)
  {
	  if (strcmp(argv[i], "-d") == 0 ||
	      strcmp(argv[i], "-display") == 0 ||
	      strcmp(argv[i], "--display") == 0)
	  {
		  if (++i >= argc)
		  {
			  usage(0);
			  exit(1);
		  }
		  display_name = argv[i];
	  }
	  else if (strcmp(argv[i], "-f") == 0)
	  {
		  flushFile = True;
	  }
	  else if (strcmp(argv[i], "-h") == 0)
	  {
		  usage(1);
		  exit(0);
	  }
	  else if (strcmp(argv[i], "-u") == 0)
	  {
		  Development = True;
	  }
	  else if (strcmp(argv[i], "-m") == 0)
	  {
		   trackMouseMotion = False;
	  }
	  else if (strcmp(argv[i], "-p") == 0)
	  {
		  if (++i >= argc)
		  {
			  usage(0);
			  exit(1);
		  }
		  privacyLevel = atoi(argv[i]);
	  }
	  else
	  {
		  usage(0);
		  fprintf(stderr, "%s: invalid option -- %s\n", myName, argv[i]);
		  exit(1);
	  }
  }

  init_sender();

  /* install an exit handler */
  sigemptyset(&sigact.sa_mask);
#ifdef SA_INTERRUPT
  sigact.sa_flags = SA_INTERRUPT;
#else
  sigact.sa_flags = 0;
#endif
  sigaddset(&sigact.sa_mask, SIGPIPE);
  sigaddset(&sigact.sa_mask, SIGTERM);
  sigaddset(&sigact.sa_mask, SIGINT);
  sigaddset(&sigact.sa_mask, SIGHUP);
  sigaddset(&sigact.sa_mask, SIGQUIT);
  sigact.sa_handler = exit_wmwatch;
  sigaction(SIGPIPE, &sigact, NULL);
  sigaction(SIGTERM, &sigact, NULL);
  sigaction(SIGINT,  &sigact, NULL);
  sigaction(SIGHUP,  &sigact, NULL);
  sigaction(SIGQUIT, &sigact, NULL);

  if (! (disp = XOpenDisplay(display_name))) {
    fprintf(stderr, "%s: Error -- Cannot open display %s\n", argv[0],
	    XDisplayName(display_name));
    return EXIT_FAILURE;
  }

  XSetErrorHandler(myErrorHandler);
  XSetIOErrorHandler(xioerror);

  atom_init0(disp);

  Window root = DefaultRootWindow(disp);

  /* now wait and check for an EWMH or GNOME1 atom */
  {
	  Bool done = False;
	  /* use usec 1 sec = 1000 ms = 1000000 usec */
	  unsigned long time_out = 120 * 1000000; /* 120 sec */
	  unsigned long partial_time_out = 2 * 1000000; /* 2 sec */
	  unsigned long time_sleep = 100 * 1000; /* 100 ms */
	  unsigned long elapsed = 0;
	  unsigned long partial_elapsed = 0;

	  while(!done)
	  {
		  if (!ewmh_compliant && get_net_check_window(disp))
		  {
			  ewmh_compliant = True;
			  XSync(disp,0);
		  }
		  if (!win_compliant && get_win_check_window(disp))
		  {
			  win_compliant = True;
			  XSync(disp,0);
			  partial_time_out = 30 * 1000000;
		  }
		  if ((ewmh_compliant && win_compliant) ||
		      elapsed >= time_out || partial_elapsed >= partial_time_out)
		  {
			  done = True;
		  }
		  else if (!ewmh_compliant && !win_compliant)
		  {
			  usleep(time_sleep);
			  elapsed += time_sleep;
		  }
		  else /* (ewmh_compliant && !win_compliant) ||
			  (ewmh_compliant && !win_compliant) */
		  {
			  usleep(time_sleep);
			  partial_elapsed += time_sleep;
			  elapsed += time_sleep;
		  }
	  }
  }

  if (!ewmh_compliant && !win_compliant)
  {
	  fprintf(stderr,"%s: Error -- The WM is neither EWMH compliant nor WIN "
		  "compliant!\n  ...exiting", argv[0]);
	  /* FIXME: send an error msg to the server */
	  send_anystr("Error", "The WM is neither EWMH compliant nor WIN");
	  do_exit = True;
  }
  else if (!ewmh_compliant)
  {
	  fprintf(stderr,"%s: ERROR -- The WM is not EWMH compliant\n",
		  argv[0]);
	  xinput_mask |= FocusChangeMask;
	  send_anystr("Error", "The WM is not EWMH compliant");
#if IS_WMTRACE
	  do_exit = True;
#endif
  }

  {
	  Bool done = False;
	  unsigned long time_out = 120 * 1000000; /* 120 sec */
	  unsigned long time_sleep = 100 * 1000; /* 100 ms */
	  unsigned long elapsed = 0;

	  while(!done && elapsed < time_out)
	  {
		  if (!(done = set_wm_info(
				CurrentTime, ewmh_compliant, win_compliant)))
		  {
			  /* should not happen any way */
			  usleep(time_sleep);
			  elapsed += time_sleep;
		  }
	  }
  }

  atom_init(disp, True);
  atom_selection.atom = XInternAtom(disp, "SELECT_WATCH", False);

  add_the_root_window();
  /* should select PointerMotionMask for XRecord (sic) */
  XSelectInput(
	  disp, root,
	  SubstructureNotifyMask | PropertyChangeMask | PointerMotionMask |
	  KeyPressMask | KeyReleaseMask);

#if IS_WMTRACE
  update_shortcut_modifiers(disp);
#endif

  set_number_of_desktops(CurrentTime);
  set_desktop_geometry(CurrentTime);
  set_desktop_viewport(CurrentTime);
  set_current_desktop(CurrentTime);
  set_desktop_names(CurrentTime);
  set_desktop_layout(CurrentTime);
  set_workarea(CurrentTime);
  set_xinerama_info(CurrentTime);
  set_virtual_roots(CurrentTime);
  set_showing_desktop(CurrentTime);

  check_xwindow(CurrentTime);
#if IS_WMTRACE
  check_cached_xwindow(CurrentTime, root);
  set_fake_button_event(CurrentTime, root);
#endif
  set_stacking_list(CurrentTime);
  set_active_xwindow(CurrentTime);

  win_select = XCreateSimpleWindow(disp, root, 0, 0, 1, 1, 0, 0, 0);
  if (!FRecordInit(disp))
  {
	  fprintf(stderr,
		  "%s: ERROR -- No XRecord support! add \"load record\" in "
		  "your X config file\n", argv[0]);
	  send_anystr("Error", "No XRecord support");
#if IS_WMTRACE
	  do_exit = True;
#endif
  }
  else
  {
	  frd = FRecordCreateDescription(disp);
	  frd->dispatch = record_dispatch;
	  /* see the event name in X.h */
#if IS_WMTRACE
	  FRecordAddRange(
		  frd, FRECORD_DELIVERED_EVENT, ButtonPress, ButtonRelease);
	  FRecordAddRange(
		  frd, FRECORD_DELIVERED_EVENT, KeyPress, KeyRelease);
	  if (trackMouseMotion)
	  {
		  FRecordAddRange(
			  frd, FRECORD_DELIVERED_EVENT, MotionNotify,
			  MotionNotify);
	  }
#endif
	  FRecordAddRange(
		  frd, FRECORD_DELIVERED_EVENT, SelectionRequest,
		  SelectionRequest);
	  FRecordStartRecording(frd);
  }


  x_fd = XConnectionNumber(disp);
  fd_width = x_fd+1;

  send_simple_msg(EndInit, ROOT_ID, CurrentTime);

  if (do_exit)
  {
	  exit_wmwatch();
  }

  while(1)
  {
	  fd_set in_fdset;
	  struct timeval timeout;
	  int num_fd = 0;

	  if (XPending(disp))
	  {
		  //FRecordProcessRepliesAll();
		  process_event();
		  //FRecordProcessRepliesAll();
	  }

	  FRecordProcessRepliesAll();

	  timeout.tv_sec = 0;
	  timeout.tv_usec = 1000 * 10; /* 10 ms */

	  FD_ZERO(&in_fdset);
	  FD_SET(x_fd, &in_fdset);

	  num_fd = select(fd_width, &in_fdset, 0, 0, &timeout);
	  if (num_fd == -1 && errno == EBADF)
	  {
		  exit_wmwatch();
	  }

	  FRecordProcessRepliesAll();

	  /* FIXME: how often we need to call XGetSelectionOwner? */
	  if (xgso_count > 5 && XGetSelectionOwner(disp, XA_PRIMARY) != None)
	  {
		  xgso_count = 0;
		  XConvertSelection(
			  disp, XA_PRIMARY, XA_STRING,
			  atom_selection.atom, win_select, CurrentTime);
	  }
	  else
	  {
		  xgso_count++;
	  }
  }

  return 0;
}
