/* -*-c-*- */
/* Copyright Olivier Chapuis (CNRS), Matthieu Langet (CNRS), 2005.
 *
 * <chapuis@lri.fr> <mathieu.langet@lri.fr>
 *
 * This software is a computer program whose purpose is to log the
 * window management activity into some files.
 *
 * This software is governed by the CeCILL  license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 *
 */


#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <sys/stat.h>
#ifdef HAVE_GETPWUID
#  include <pwd.h>
#endif

#include <X11/Xlib.h>
#include <X11/keysym.h>

#include "flist.h"
#include "global.h"
#include "xget.h"
#include "xwindow.h"
#include "trace.h"

#define TRACE_FORMAT_VERSION "1.0"

static FILE *file = NULL;
static char *wmtrace_file;
static char *wmtrace_file_archive;
static char *wmtrace_dir_archive;
static time_t start_time;
static struct timeval startTime;


extern char *myName;
extern Bool Development;
extern Bool flushFile;
extern int privacyLevel;

extern xwindow list_xwindow[1024];
extern int list_xwindow_size;

typedef struct
{
	char *type;
	unsigned long uid;
	char *msg;
	long time;
	int event_time;
	button_event *be;
} trace_msg_t;

#define TRACE_MSG_INIT(x) \
   trace_msg_t *tmt; \
   char *msg; \
   if (x > 0) msg = malloc(x*sizeof(char)); \
   tmt = (trace_msg_t *)malloc(sizeof(trace_msg_t)); \
   tmt->be = NULL;

#define TRACE_MSG_SET(ty,id,m,t,s) \
   tmt->type = ty; tmt->uid = id; \
   tmt->msg = m; \
   tmt->time = t; tmt->event_time = s; tmt->be = NULL; \
   check_and_send(tmt);

#define TRACE_MSG_SET_BE(t,s,e) \
   tmt->msg = NULL; \
   tmt->time = t; tmt->event_time = s; \
   tmt->be = (button_event *)malloc(sizeof(button_event)); \
   memcpy(tmt->be, e, sizeof(button_event)); \
   check_and_send(tmt);

#define TRACE_MSG_FREE(tmt) \
  if (tmt != NULL) { \
    if (tmt->msg) free(tmt->msg); \
    if (tmt->be) free(tmt->be); \
    free(tmt); \
  }

static flist *msg_queue = NULL;

static
void my_flush(void)
{
	if (Development || flushFile)
	{
		fflush(file);
	}
}

static
void print_trace_msg(trace_msg_t *ctmt)
{
	if (ctmt->be != NULL)
	{
		fprintf(file, "<ButtonMotion uid='%lu' t='%lu' x='%d' y='%d' "
			"button='%d' state='%ld'/>\n",
			ctmt->be->uid, ctmt->time, ctmt->be->x, ctmt->be->y,
			ctmt->be->button, ctmt->be->state);
	}
	else
	{
		fprintf(file, "<%s uid='%lu' t='%lu'%s", ctmt->type, ctmt->uid,
			ctmt->time, ctmt->msg);
	}
}

static 
Bool flush_msg()
{
	flist *tl;
	trace_msg_t *ctmt;
	trace_msg_t *ntmt;
	Bool printed = False;
	unsigned long start_time;
	
	if (msg_queue == NULL)
	{
		return False;
	}
	
	tl = msg_queue;
	ctmt = (trace_msg_t *)(tl->object);
	fprintf(file, "<ButtonMotionSeq uid='%lu' t='%lu' button='%d' "
		"state='%ld'/>", ctmt->be->uid, ctmt->time, ctmt->be->button,
		ctmt->be->state);
	start_time = ctmt->time;

	while(tl)
	{
		ctmt = (trace_msg_t *)(tl->object);
		if (tl->next != NULL)
		{
			ntmt = (trace_msg_t *)(tl->next->object);
		}
		else
		{
			ntmt = NULL;
		}
		if (ntmt != NULL && ntmt->time <= ctmt->time + 1)
		{
			/* skip it */
		}
		else
		{
			fprintf(file, " %lu %d %d",
				ctmt->time - start_time, ctmt->be->x,
				ctmt->be->y);
		}
		msg_queue = flist_remove_obj(msg_queue, tl->object);
		TRACE_MSG_FREE(ctmt);
		tl = msg_queue;
		printed = True;
	}
	fprintf(file, "</ButtonMotionSeq>\n");
	
	return printed;
}

static 
void add_to_flush_queue(trace_msg_t *tmt)
{
	msg_queue = flist_append_obj(msg_queue, tmt);
}


static
void check_and_send(trace_msg_t *tmt)
{
	static button_event *current_be = NULL;

	if (tmt->be == NULL || tmt->be->type != wmButtonMotion ||
	    (current_be != NULL &&
	     (current_be->uid != tmt->be->uid ||
	      current_be->button != tmt->be->button ||
	      current_be->state != tmt->be->state)))
	{
		flush_msg(0);
		if (tmt->be != NULL)
		{
			current_be = tmt->be;
			add_to_flush_queue(tmt);
		}
		else
		{
			print_trace_msg(tmt);
			TRACE_MSG_FREE(tmt);
			current_be = NULL;
		}
		my_flush();
	}
	else
	{
		current_be = tmt->be;
		add_to_flush_queue(tmt);
	}
}

static
int timeval_subtract (
	struct timeval *result, struct timeval *x, struct timeval *y)
{
	/* Perform the carry for the later subtraction by updating Y. */
	if (x->tv_usec < y->tv_usec)
	{
		int nsec = (y->tv_usec - x->tv_usec) / 1000000 + 1;
		y->tv_usec -= 1000000 * nsec;
		y->tv_sec += nsec;
	}
	if (x->tv_usec - y->tv_usec > 1000000)
	{
		int nsec = (x->tv_usec - y->tv_usec) / 1000000;
		y->tv_usec += 1000000 * nsec;
		y->tv_sec -= nsec;
	}
     
	/* Compute the time remaining to wait.
	   `tv_usec' is certainly positive. */
	result->tv_sec = x->tv_sec - y->tv_sec;
	result->tv_usec = x->tv_usec - y->tv_usec;
     
	/* Return 1 if result is negative. */
	return x->tv_sec < y->tv_sec;
}

static
unsigned long get_time(void)
{
	struct timeval ct;
	struct timeval t;

	gettimeofday (&ct, NULL);
	timeval_subtract (&t, &ct, &startTime);
	return (1000*t.tv_sec + t.tv_usec/1000);
}

static
char *action_string(int id)
{
	char *data;
  
	switch (id)
	{
	case CreateWindow:
		data = "CreateWindow";
		break;
	case DestroyWindow:
		data = "DestroyWindow";
		break;
	case ResizeWindow:
		data = "ResizeWindow";
		break;
	case ChangeNameWindow:
		data = "ChangeNameWindow";
		break;
	case IconGeometry:
		data = "IconGeometry";
		break;
	case WindowStrut:
		data = "WindowStrut";
		break;
	case FrameExtents:
		data = "FrameExtents";
		break;
	case PartialWindowStrut:
		data = "PartialWindowStrut";
		break;
	case HandleIcons:
		data = "HandleIcons";
		break;
	case AllowedActions:
		data = "AllowedActions";
		break;
	case MaximizeWindow:
		data = "MaximizeWindow";
		break;
	case UnMaximizeWindow:
		data = "UnMaximizeWindow";
		break;
	case MinimizeWindow:
		data = "MinimizeWindow";
		break;
	case UnMinimizeWindow:
		data = "UnMinimizeWindow";
		break;
	case HideWindow:
		data = "HideWindow";
		break;
	case UnHideWindow:
		data = "UnHideWindow";
		break;
	case ShadeWindow:
		data = "ShadeWindow";
		break;
	case UnShadeWindow:
		data = "UnShadeWindow";
		break;
	case SkipTaskBarWindow:
		data = "SkipTaskBarWindow";
		break;
	case UnSkipTaskBarWindow:
		data = "UnSkipTaskBarWindow";
		break;
	case SkipPagerWindow:
		data = "SkipPagerWindow";
		break;
	case UnSkipPagerWindow:
		data = "UnSkipPagerWindow";
		break;
	case FullScreenWindow:
		data = "FullScreenWindow";
		break;
	case UnFullScreenWindow:
		data = "UnFullScreenWindow";
		break;
	case ModalWindow:
		data = "ModalWindow";
		break;
	case UnModalWindow:
		data = "UnModalWindow";
		break;	
	case TakeFocus:
		data = "TakeFocus";
		break;
	case ReleaseFocus:
		data = "ReleaseFocus";
		break;
	case ActiveWindow:
		data = "ActiveWindow";
		break;
	case MoveWindow:
		data = "MoveWindow";
		break;
	case MoveResizeWindow:
		data = "MoveResizeWindow";
		break;
	case AlwaysVisibleWindow:
		data = "StickWindow";
		break;
	case NotAlwaysVisibleWindow:
		data = "UnStickWindow";
		break;
	case LayerChange:
		data = "LayerChange";
		break;	
	case ChangeDesktopWindow:
		data = "ChangeDesktopWindow";
		break;
	case ChangeCurrentDesktop:
		data = "ChangeCurrentDesktop";
		break;
	case ClientStackingList:
		data = "ClientStackingList";
		break;
	case NumberOfDesktops:
		data = "NumberOfDesktops";
		break;
	case DesktopViewport:
		data = "DesktopViewport";
		break;
	case DesktopGeometry:
		data = "DesktopGeometry";
		break;
	case DesktopNames:
		data = "DesktopNames";
		break;
	case DesktopLayout:
		data = "DesktopLayout";
		break;
	case DesktopWorkarea:
		data = "DesktopWorkarea";
		break;
	case ShowingDesktop:
		data = "ShowingDesktop";
		break;
	case WindowManagerInfo:
		data = "WindowManagerInfo";
		break;
	case XineramaInfo:
		data = "XineramaInfo";
		break;
	case wmButtonPress:
		data = "ButtonPress";
		break;
	case wmButtonRelease:
		data = "ButtonRelease";
		break;
	case wmButtonMotion:
		data = "ButtonMotion";
		break;
	case wmKeyPress:
		data = "KeyPress";
		break;
	case wmKeyRelease:
		data = "KeyRelease";
		break;	
	case wmSelectionNotify:
		data = "SelectionNotify";
		break;
	case wmSelectionRequest:
		data = "SelectionRequest";
		break;
	case EndInit:
		data = "EndInit";
		break;
	default:
		data = NULL;
		break;
	}
	return data;
}

typedef struct
{
	char key;
	int  value;
	char name;
} charmap_t;


static charmap_t key_modifiers[] =
{
	{'S', ShiftMask, 'S'},
	{'C', ControlMask, 'C'},
	{'L', LockMask, 'N'},
	{'M', Mod1Mask, '0'},
	{'2', Mod2Mask, '0'},
	{'3', Mod3Mask, '0'},
	{'4', Mod4Mask, '0'},
	{'5', Mod5Mask, '0'},
};

static
char *modifiers_string(int m)
{
	static char s[256];
	int index = 0;
	int i = 0;

	for(i = 0; i < 8; i++)
	{
		if ((m & key_modifiers[i].value) && key_modifiers[i].name != '0')
		{
			s[index++] = key_modifiers[i].name;
		}
	}
	s[index++] = '\0';

	return s;
}

static int shortcut_modifiers = 0;

static
Bool is_shortcut_modifiers(int m)
{
	if (m & shortcut_modifiers)
	{
		return True;
	}

	return False;
}

static
Bool is_modifier_key(KeySym k)
{
	switch(k)
	{
#if defined(XK_Meta_L) && defined(XK_Meta_R)
	case XK_Meta_L:
	case XK_Meta_R:
#endif
#if defined(XK_Alt_L) && defined(XK_Alt_R)
	case XK_Alt_L:
	case XK_Alt_R:
#endif
#if defined(XK_Super_L) && defined(XK_Super_R)
	case XK_Super_L:
	case XK_Super_R:
#endif
#if defined(XK_Hyper_L) && defined(XK_Hyper_R)
	case XK_Hyper_L:
	case XK_Hyper_R:
#endif
#if defined(XK_Shift_L) && defined(XK_Shift_R)
	case XK_Shift_L:
	case XK_Shift_R:
#endif
#if defined(XK_Control_L) && defined(XK_Control_R)
	case XK_Control_L:
	case XK_Control_R:
#endif
#if 0
	case XK_Num_Lock:
	case XK_Shift_Lock:
	case XK_Caps_Lock:
#endif
		return True;
		break;
	default:
		break;
	}
	return False;
}

/* ------------------------ Interface functions ------------------------------ */

/* basically we want to found the Super key ... */
void update_shortcut_modifiers(Display *disp)
{
	int j;
	XModifierKeymap *mm;

	mm = XGetModifierMapping(disp);

	shortcut_modifiers |= Mod1Mask|ControlMask;

	if (mm == NULL)
	{
		return;
	}

	for(j = 0; j < 8*mm->max_keypermod; j = j + mm->max_keypermod)
	{
		KeySym k = XKeycodeToKeysym(disp, mm->modifiermap[j],0);
		switch(k)
		{
#if defined(XK_Meta_L) && defined(XK_Meta_R)
		case XK_Meta_L:
		case XK_Meta_R:
			shortcut_modifiers |=
				key_modifiers[j/mm->max_keypermod].value;
			key_modifiers[j/mm->max_keypermod].name = 'M';
			break;
#endif
#if defined(XK_Alt_L) && defined(XK_Alt_R)
		case XK_Alt_L:
		case XK_Alt_R:
			shortcut_modifiers |=
				key_modifiers[j/mm->max_keypermod].value;
			key_modifiers[j/mm->max_keypermod].name = 'A';
			break;
#endif
#if defined(XK_Super_L) && defined(XK_Super_R)
		case XK_Super_L:
		case XK_Super_R:
			shortcut_modifiers |=
				key_modifiers[j/mm->max_keypermod].value;
			key_modifiers[j/mm->max_keypermod].name = 'W';
			break;
#endif
#if defined(XK_Hyper_L) && defined(XK_Hyper_R)
		case XK_Hyper_L:
		case XK_Hyper_R:
			shortcut_modifiers |=
				key_modifiers[j/mm->max_keypermod].value;
			key_modifiers[j/mm->max_keypermod].name = 'H';
			break;
#endif
#if defined(XK_Shift_L) && defined(XK_Shift_R)
		case XK_Shift_L:
		case XK_Shift_R:
			key_modifiers[j/mm->max_keypermod].name = 'S';
			break;
#endif
#if defined(XK_Control_L) && defined(XK_Control_R)
		case XK_Control_L:
		case XK_Control_R:
			shortcut_modifiers |=
				key_modifiers[j/mm->max_keypermod].value;
			key_modifiers[j/mm->max_keypermod].name = 'C';
			break;
#endif
#if 0
		case XK_Num_Lock:
			key_modifiers[j/mm->max_keypermod].name = 'N';
			break;
		case XK_Shift_Lock:
			key_modifiers[j/mm->max_keypermod].name = 'L';
			break;
		case XK_Caps_Lock:
			key_modifiers[j/mm->max_keypermod].name = 'K';
			break;
#endif
		default:
			break;
		}
	}
	XFreeModifiermap(mm);
}

void close_sender(void)
{
	unsigned long t;
	char *cmd;
	TRACE_MSG_INIT(32);

	t = get_time();
	sprintf(msg, "/>\n</Session>\n");
	TRACE_MSG_SET("EndSession", 0, msg, t, CurrentTime);

	fflush(file);
	fclose(file);

	if (Development)
	{
		return;
	}

	if (rename(wmtrace_file, wmtrace_file_archive) == -1)
	{
		fprintf(stderr, "[%s] fail to rename %s to %s\n", myName,
			wmtrace_file, wmtrace_file_archive);
		return;
	}

	if (chdir(wmtrace_dir_archive) == -1)
	{
		fprintf(stderr, "[%s] fail to chdir %s\n", myName,
			wmtrace_dir_archive);	
	}
	cmd = malloc(strlen(wmtrace_file_archive) + 12);
	sprintf(cmd, "bzip2 %s", wmtrace_file_archive);
		     
	if (system(cmd) == -1)
	{
		fprintf(stderr, "[%s] fail to bzip2 %s\n", myName,
			wmtrace_file_archive);
		return;	
	}

	/* send a mail */
}

int init_sender(void)
{
	char *home_dir;
	char *user;
	char *wmtrace_dir;
	struct tm *broken_down_time;
	char date_string[256];
	unsigned long t;

	home_dir = getenv("HOME");
	user = getenv("USER");

#ifdef HAVE_GETPWUID
	if (home_dir == NULL)
	{
		struct passwd* pw = getpwuid(getuid());
		if (pw != NULL)
		{
			home_dir = strdup(pw->pw_dir);
		}
	}
#endif
	if (home_dir == NULL)
	{
		home_dir = "/tmp"; /* give up */
	}

	wmtrace_dir = (char *)malloc(strlen(home_dir) + 12); 
	sprintf(wmtrace_dir, "%s/.wmtrace", home_dir);

	if (access(wmtrace_dir, F_OK) != 0)
	{
		mkdir(wmtrace_dir, 0777);
	}
	if (access(wmtrace_dir, W_OK) != 0)
	{
		fprintf(stderr, "[%s] Fail to create directory %s\n", myName,
			wmtrace_dir);
		exit(1);
	}

	chdir(wmtrace_dir);

	wmtrace_dir_archive = (char *)malloc(strlen(wmtrace_dir) + 12); 
	sprintf(wmtrace_dir_archive, "%s/archive", wmtrace_dir);

	if (access(wmtrace_dir_archive, F_OK) != 0)
	{
		mkdir(wmtrace_dir_archive, 0777);
	}
	if (access(wmtrace_dir_archive, W_OK) != 0)
	{
		fprintf(stderr, "[%s] Fail to create directory %s\n",
			myName, wmtrace_dir_archive);
		exit(1);
	}

	time(&start_time);
	gettimeofday (&startTime, NULL);
	broken_down_time = localtime(&start_time);

	sprintf(date_string,"%i-%s%i-%s%i-%s%d:%s%d:%s%d",
		broken_down_time->tm_year + 1900,
		(broken_down_time->tm_mon+1 < 10)? "0":"",
		broken_down_time->tm_mon+1,
		(broken_down_time->tm_mday < 10)? "0":"",
		broken_down_time->tm_mday,
		(broken_down_time->tm_hour < 10)? "0":"",
		broken_down_time->tm_hour,
		(broken_down_time->tm_min < 10)? "0":"",
		broken_down_time->tm_min,
		(broken_down_time->tm_sec < 10)? "0":"",
		broken_down_time->tm_sec);

	wmtrace_file = (char *)malloc(
		strlen(wmtrace_dir) + strlen(date_string) + 24);
	wmtrace_file_archive = (char *)malloc(
		strlen(wmtrace_dir_archive) + strlen(date_string) + 24);

	if (Development)
	{
		sprintf(wmtrace_file, "%s/trace-devel", wmtrace_dir);
	}
	else
	{
		sprintf(wmtrace_file, "%s/trace-%s", wmtrace_dir,date_string);
	}
	sprintf(wmtrace_file_archive, "%s/trace-%s", wmtrace_dir_archive,
		date_string);

	if ((file = fopen(wmtrace_file, "a")) == NULL)
	{
		fprintf(stderr, "[%s] Fail to create file %s\n", myName,
			wmtrace_file);
		exit(1);
	}

	t = startTime.tv_sec*1000 + startTime.tv_usec/1000;

	{
		int len;
		char *user_str;
		TRACE_MSG_INIT(0);
		
		user_str = (user)? user:home_dir;
		len = 26 + strlen(user_str) + strlen(VERSION) +
			strlen(TRACE_FORMAT_VERSION);
		msg = malloc(len);
		sprintf(msg, " user='%s' v='%s' f='%s'>\n",
			user_str, VERSION, TRACE_FORMAT_VERSION);
		TRACE_MSG_SET("Session", 0, msg, t, CurrentTime);
	}

	free(wmtrace_dir);
	return fileno(file);
}

void send_simple_msg(int type, int id, int event_time)
{
	char *action_str;
	unsigned long t;

	action_str = action_string(type);
	if (action_str == NULL)
	{
		/* we do not want to handle such msg */
		return;
	}
	
	{
		TRACE_MSG_INIT(8);

		t = get_time();
		sprintf(msg,"/>\n");
		TRACE_MSG_SET(
			action_str, list_xwindow[id].uid_, msg, t, event_time);
	}
}

void send_long_msg(int type, int id, int event_time, long k)
{
	char *action_str;
	char *key = "";

	action_str = action_string(type);
	if (action_str == NULL)
	{
		/* we do not want to handle such msg */
		return;
	}

	switch(type)
	{
	case AllowedActions:
		key = "allowed_actions";
		break;
	default:
		break;
	}

	{
		unsigned long t;
		TRACE_MSG_INIT(64);

		t = get_time();
		sprintf(msg, " %s='%ld'/>\n", key, k);
		TRACE_MSG_SET(
			action_str, list_xwindow[id].uid_, msg, t, event_time);
	}	
}

void send_unsigned_msg(int type, int id, int event_time, unsigned long k)
{
	char *action_str;
	char *key = "";

	action_str = action_string(type);
	if (action_str == NULL)
	{
		/* we do not want to handle such msg */
		return;
	}

	switch(type)
	{
	case NumberOfDesktops:
	case ChangeCurrentDesktop:
	case ChangeDesktopWindow:
		key = "desk";
		break;
	case  LayerChange:
		key = "layer";
		break;
	case ShowingDesktop:
	case HandleIcons:
		key = "on";
		break;
	default:
		break;
	}

	{
		unsigned long t;
		TRACE_MSG_INIT(64);

		t = get_time();
		sprintf(msg," %s='%lu'/>\n", key, k);
		TRACE_MSG_SET(
			action_str, list_xwindow[id].uid_, msg, t, event_time);
	}
}

void send_size(
	int type, int id, int event_time, unsigned long w, unsigned long h)
{
	char *action_str;

	action_str = action_string(type);
	if (action_str == NULL)
	{
		/* we do not want to handle such msg */
		return;
	}

	{
		unsigned long t;
		TRACE_MSG_INIT(64);

		t = get_time();
		sprintf(msg, " w='%lu' h='%lu'/>\n", w, h);
		TRACE_MSG_SET(
			action_str, list_xwindow[id].uid_, msg, t, event_time);
	}
}

void send_position(int type, int id, int event_time, long x, long y)
{
	char *action_str;

	action_str = action_string(type);
	if (action_str == NULL)
	{
		/* we do not want to handle such msg */
		return;
	}

	{
		unsigned long t;
		TRACE_MSG_INIT(64);

		t = get_time();
		sprintf(msg, " x='%ld' y='%ld'/>\n", x, y);
		TRACE_MSG_SET(
			action_str, list_xwindow[id].uid_, msg, t, event_time);
	}
}

void send_geometry(int type, int id, int event_time, int x, int y, int w, int h)
{
	char *action_str;

	action_str = action_string(type);
	if (action_str == NULL)
	{
		/* we do not want to handle such msg */
		return;
	}

	{
		unsigned long t;
		TRACE_MSG_INIT(128);

		t = get_time();
		sprintf(msg,
			" x='%d' y='%d' w='%d' h='%d'/>\n", x, y, w, h);
		TRACE_MSG_SET(
			action_str, list_xwindow[id].uid_, msg, t, event_time);
	}
}

void send_name(int type, int id, int event_time, char *str, int net)
{
	unsigned long t;
	char *action_str;

	action_str = action_string(type);
	if (action_str == NULL)
	{
		/* we do not want to handle such msg */
		return;
	}
	
	{
		char *my_str;
		TRACE_MSG_INIT(0);

		t = get_time();
		my_str = (privacyLevel > 0)? "Private":str;
		if (my_str == NULL)
		{
			my_str = "";
		}
		msg = malloc(
			(32 + strlen(action_str) + strlen(my_str))*sizeof(char));
		sprintf(msg," utf8='%i'>%s</%s>\n", net, my_str, action_str);
		TRACE_MSG_SET(
			action_str, list_xwindow[id].uid_, msg, t, event_time);
	}
}

void send_names_list(
	int type, int id, int event_time, unsigned char **names,
	unsigned int nbr)
{
	char *action_str;

	action_str = action_string(type);
	if (action_str == NULL)
	{
		/* we do not want to handle such msg */
		return;
	}
	
	{
		int l,i;
		unsigned long t;
		char tmp[256];
		TRACE_MSG_INIT(0);

		t = get_time();
		l = 32;
		for (i = 0; i < nbr; i++)
		{
			if (names[i] != NULL)
			{
				l = l + strlen(names[i]);
			}
			l = l + 21 + 12;
		}

		msg = malloc(l*sizeof(char));
		sprintf(msg," num='%i'>", nbr);
		for (i = 0; i < nbr; i++)
		{
			if (names[i] != NULL)
			{
				sprintf(tmp,"<name num='%d'>", i);
				strcat(msg,tmp);
				strcat(msg,names[i]);
				strcat(msg,"</name>");
			}
			else
			{
				sprintf(tmp,
					"<name num='%d'></name>", i);
				strcat(msg,tmp);
			}
		}
		
		sprintf(tmp,"</%s>\n", action_str);
		strcat(msg,tmp);
		TRACE_MSG_SET(
			action_str, list_xwindow[id].uid_, msg, t, event_time);
	}
}

void send_desktop_layout(int event_time, desktop_layout_t *dl)
{
	unsigned long t;
	TRACE_MSG_INIT(256);

	t = get_time();
	sprintf(msg," orientation='%lu' "
		"columns='%lu' rows='%lu' sc='%lu'/>\n",
		dl->orientation, dl->columns, dl->rows, dl->starting_corner);
	TRACE_MSG_SET("DesktopLayout", 0, msg, t, event_time);
}

void send_strut(int type, int id, int event_time, strut_t *s)
{
	char *action_str;
	
	action_str = action_string(type);
	if (action_str == NULL)
	{
		/* we do not want to handle such msg */
		return;
	}
	
	{
		unsigned long t;
		TRACE_MSG_INIT(128);

		t = get_time();	
		if (s == NULL)
		{
			sprintf(msg,"/>\n");
		}
		else
		{
			sprintf(msg,
				" left='%ld' right='%ld' top='%ld' "
				"bottom='%ld'/>\n",
				s->left, s->right, s->top, s->bottom);
		}
		TRACE_MSG_SET(
			action_str, list_xwindow[id].uid_, msg, t, event_time);
	}
}

void send_partial_strut(int type, int id, int event_time, partial_strut_t *s)
{
	char *action_str;

	action_str = action_string(type);
	if (action_str == NULL)
	{
		/* we do not want to handle such msg */
		return;
	}
	
	{
		unsigned long t;
		TRACE_MSG_INIT(512);

		t = get_time();
		if (s == NULL)
		{
			sprintf(msg,"/>\n");
		}
		else
		{
			sprintf(msg," pleft='%ld' "
				"pright='%ld' ptop='%ld' pbottom='%ld' "
				"pleft_start_y='%ld' pleft_end_y='%ld' "
				"pright_start_y='%ld' pright_end_y='%ld' "
				"ptop_start_x='%ld' ptop_end_y='%ld' "
				"pbottom_start_x='%ld' pbottom_end_y='%ld'/>",
				s->left, s->right, s->top, s->top,
				s->start.left, s->end.left,
				s->start.right, s->end.right,
				s->start.top, s->end.top,
				s->start.bottom, s->end.bottom);
		}
		TRACE_MSG_SET(
			action_str, list_xwindow[id].uid_, msg, t, event_time);
	}
	
}

void send_button_msg(int event_time, button_event *be)
{
	char *action_str;

	action_str = action_string(be->type);
	if (action_str == NULL)
	{
		/* we do not want to handle such msg */
		return;
	}
	
	{
		unsigned long t;
		t = get_time();

		if (be->type != wmButtonMotion)
		{
			TRACE_MSG_INIT(128);

			sprintf(msg," x='%d' y='%d' button='%d' state='%ld'/>\n",
				be->x, be->y, be->button, be->state);
			TRACE_MSG_SET(
				action_str, be->uid, msg, t,
				event_time);
		}
		else
		{
			TRACE_MSG_INIT(0);
			TRACE_MSG_SET_BE(t, event_time, be);
		}
	}
	
}

void send_key_msg(int type, int id, int event_time, int key, int state)
{
	char *action_str;
	static char *prev_key = NULL;
	static char *prev_mod_str = NULL;

	action_str = action_string(type);
	if (action_str == NULL)
	{
		/* we do not want to handle such msg */
		return;
	}

	if (is_modifier_key(key))
	{
		return;
	}
	
	{
		unsigned long t;
		char *key_string;
		char *modifier_str;
		Bool private = False;
		Bool same = False;
		TRACE_MSG_INIT(256);

		t = get_time();

		if (type == wmKeyRelease)
		{
			sprintf(msg,"/>\n");
			TRACE_MSG_SET(
				action_str, list_xwindow[id].uid_, msg, t,
				event_time);
			return;
		}
		key_string = XKeysymToString(key);
		modifier_str = modifiers_string(state);
		if (privacyLevel > 0)
		{
			/* detect short cut */
			if (is_shortcut_modifiers(state))
			{
				/* ok */
			}
			else
			{
				private = True;
			}
		}
		if (prev_key != NULL && prev_mod_str != NULL &&
		    key_string != NULL && modifier_str != NULL)
		{
			if (strcmp(prev_key, key_string) == 0 &&
			    strcmp(prev_mod_str, modifier_str) == 0)
			{
				same = True;
			}
		}
		
		if (!same)
		{
			if (prev_key != NULL)
			{
				free(prev_key);
				prev_key = NULL;
			}
			if (prev_mod_str != NULL)
			{
				free(prev_mod_str);
				prev_mod_str = NULL;
			}
			if (key_string != NULL && modifier_str != NULL)
			{
				prev_mod_str = strdup(modifier_str);
				prev_key = strdup(key_string);
			}
		}

		if (same)
		{
			sprintf(msg," s='1'/>\n");	
		}
		else if (key_string == NULL || private)
		{
			sprintf(msg,"/>\n");
		}
		else
		{
			sprintf(msg," key='%s' modifiers='%s'/>\n",
				key_string, modifier_str);
		}
		TRACE_MSG_SET(
			action_str, list_xwindow[id].uid_, msg, t, event_time);
	}
}

void send_selection_notify(int event_time, int owner_id, char *val)
{
	if (val == NULL || owner_id == ROOT_ID)
	{
		/* if we do not know the owner do not care */
		return;
	}

	{		
		int l;
		unsigned long t;
		TRACE_MSG_INIT(256);

		t = get_time();
		l = strlen(val);
		sprintf(msg, " length='%d'/>\n", l);
		TRACE_MSG_SET(
			"SelectionNotify", list_xwindow[owner_id].uid_, msg,
			t, event_time);
	}
}

void send_selection_request(int event_time, int owner_id, int requestor_id)
{
	if (requestor_id == ROOT_ID)
	{
		/* if we do not know the requestor do not care, this
		 * prevent repeated msg caused by a clipper app */
		return;
	}
	
	{
		unsigned long t;
		TRACE_MSG_INIT(32);

		t = get_time();
		sprintf(msg,
			" owner='%lu'/>\n", list_xwindow[owner_id].uid_);
		TRACE_MSG_SET(
			"SelectionRequest", list_xwindow[requestor_id].uid_, msg,
			t, event_time);
	}
	
}

void send_create_window(int id, int event_time)
{
	unsigned long t;
	char *action_str = "CreateWindow";
	t = get_time();
	
	if (list_xwindow[id].management_ == MM_OVERRIDE_REDIRECT)
	{
		action_str = "CreateORWindow";
	}

	if (id == ROOT_ID)
	{
		TRACE_MSG_INIT(128);
		sprintf(msg, " w='%d' h='%d' focused='%d' active='%d'/>\n",
			list_xwindow[id].w_,
			list_xwindow[id].h_,
			list_xwindow[id].focus_,
			list_xwindow[id].active_);
		TRACE_MSG_SET(
			action_str, list_xwindow[id].uid_, msg, t, event_time);
	}
	else if (list_xwindow[id].management_ == MM_OVERRIDE_REDIRECT)
	{
		TRACE_MSG_INIT(128);
		sprintf(msg," w='%d' h='%d' x='%d' y='%d'/>\n",
			list_xwindow[id].w_,
			list_xwindow[id].h_,
			list_xwindow[id].x_, list_xwindow[id].y_);
		TRACE_MSG_SET(
			action_str, list_xwindow[id].uid_, msg, t, event_time);
	}
	else
	{
		int tid = -1;
		char strut_str[256];
		char partial_strut_str[1024];
		char frame_extents_str[256];
		char transient_str[256];
		char icon_g_str[256];
		char *name_str;
		char *res_str;
		char *class_str;
		int len = 0;
		TRACE_MSG_INIT(0);

		strut_str[0] = '\0';
		frame_extents_str[0] = '\0';
		partial_strut_str[0] = '\0';
		transient_str[0]  = '\0';
		icon_g_str[0] = '\0';

		if (list_xwindow[id].transientFor_ != None &&
		    (tid = get_id_xwindow(
			    list_xwindow[id].transientFor_)) != -1)
		{
			sprintf(transient_str," transient='%lu'",
				list_xwindow[tid].uid_);
		}
		
		if (list_xwindow[id].partial_strut_)
		{
			partial_strut_t *s = list_xwindow[id].partial_strut_;
			sprintf(partial_strut_str,
				" pleft='%ld' pright='%ld' ptop='%ld' "
				"pbottom='%ld' "
				"pleft_start_y='%ld' pleft_end_y='%ld' "
				"pright_start_y='%ld' pright_end_y='%ld' "
				"ptop_start_x='%ld' ptop_end_y='%ld' "
				"pbottom_start_x='%ld' pbottom_end_y='%ld' ",
				s->left, s->right, s->top, s->top,
				s->start.left, s->end.left,
				s->start.right, s->end.right,
				s->start.top, s->end.top,
				s->start.bottom, s->end.bottom);
		}
		else if (list_xwindow[id].strut_)
		{
			strut_t *s = list_xwindow[id].strut_;
			sprintf(strut_str,
				" left='%ld' right='%ld' top='%ld' bottom='%ld'",
				s->left, s->right, s->top, s->bottom);
		}
		else if (list_xwindow[id].frame_extents_)
		{
			strut_t *s = list_xwindow[id].frame_extents_;
			sprintf(frame_extents_str,
				" fleft='%ld' fright='%ld' ftop='%ld' "
				"fbottom='%ld'",
				s->left, s->right, s->top, s->bottom);
		}
		if (list_xwindow[id].icon_w_ > 0 && list_xwindow[id].icon_h_ > 0)
		{
			
			sprintf(icon_g_str,
				" ix='%d' iy='%d' iw='%d' ih='%d'",
				list_xwindow[id].icon_x_,
				list_xwindow[id].icon_y_,
				list_xwindow[id].icon_w_,
				list_xwindow[id].icon_h_);
		}
		name_str = (privacyLevel > 0)? "Private":list_xwindow[id].name_;
		res_str = (list_xwindow[id].classhints_.res_name)?
			list_xwindow[id].classhints_.res_name : 
			"NoResource";
		
		class_str = (list_xwindow[id].classhints_.res_class)?
			list_xwindow[id].classhints_.res_class : 
			"NoClass",
		len = 512 +
			strlen(transient_str) +
			strlen(partial_strut_str) +
			strlen(strut_str) +
			strlen(icon_g_str) +
			strlen(name_str) +
			strlen(res_str) +
			strlen(class_str) +
			strlen(list_xwindow[id].type_);
		msg = malloc(len * sizeof(char));
		sprintf(msg, " x='%d' y='%d' w='%d' "
			"h='%d' focused='%d' active='%d' isNetName='%d' "
			"desk='%lu' state='%lu' layer='%d' handle_icons='%d' "
			"allowed_actions='%ld'%s%s%s%s%s>"
			"<name>%s</name>"
			"<res_name>%s</res_name>"
			"<class_name>%s</class_name>"
			"<type>%s</type></CreateWindow>\n",
			list_xwindow[id].x_, list_xwindow[id].y_,
			list_xwindow[id].w_,
			list_xwindow[id].h_,
			list_xwindow[id].focus_,
			list_xwindow[id].active_,
			list_xwindow[id].is_net_name_,
			list_xwindow[id].desktop_,
			list_xwindow[id].state_,
			list_xwindow[id].layer_,
			list_xwindow[id].handle_icons_,
			list_xwindow[id].allowed_actions_,
			transient_str, partial_strut_str, strut_str,
			frame_extents_str, icon_g_str,
			name_str, res_str, class_str,
			list_xwindow[id].type_);
		TRACE_MSG_SET(
			action_str, list_xwindow[id].uid_, msg, t, event_time);
	}
}

void send_wm_info(int event_time, char *wm_name, char *ewmh, char *win)
{
	unsigned long t;
	int len;
	TRACE_MSG_INIT(0);

	t = get_time();
	len = 256 + strlen(wm_name) +  strlen(ewmh) + strlen(win);
	msg = malloc(len * sizeof(char));
	
	sprintf(msg,
		"><name>%s</name><ewmh>%s</ewmh><win>%s</win>"
		"</WindowMangerInfo>\n", wm_name, ewmh, win);
	TRACE_MSG_SET("WindowMangerInfo", 0, msg, t, event_time);
}


void send_stacking_list(int event_time, int *list, int size)
{
	unsigned long t;
	int i;
	int len;
	char *tmp_msg;
	char tmp[32];
	TRACE_MSG_INIT(0);

	t = get_time();

	tmp_msg = malloc(size * 12 + 1);
	tmp_msg[0] = '\0';

	for(i = 0; i < size; i++)
	{
		sprintf(tmp, "%d", list[i]);
		if (i == size-1)
		{
			sprintf(tmp, "%d", list[i]);
		}
		else
		{
			sprintf(tmp, "%d ", list[i]);
		}
		strcat(tmp_msg,tmp);
	}

	len = 32 + strlen(tmp_msg);
	msg = malloc(len * sizeof(char));
	sprintf(msg, ">%s</ClientStackingList>\n", tmp_msg);
	TRACE_MSG_SET("ClientStackingList", 0, msg, t, event_time);
	free(tmp_msg);
}

void send_xinerama_info(int event_time, XineramaScreenInfo *info, int count)
{
	unsigned long t;
	int i;
	char tmp[256];
	TRACE_MSG_INIT(0);

	t = get_time();
	msg = malloc(16 + 128*count + 25);
	sprintf(msg, " count='%d'>", count);
	for(i = 0; i < count; i++)
	{
		sprintf(tmp, "<Screen num='%d' x='%d' y='%d' w='%d' h='%d'/>",
			i, info[i].x_org, info[i].y_org, info[i].width,
			info[i].height);
		strcat(msg,tmp);
	}
	strcat(msg,"</XineramaInfo>\n");
	TRACE_MSG_SET("XineramaInfo", 0, msg, t, event_time);
}

void send_xerror(XErrorEvent *e)
{
	unsigned long t;
	t = get_time();

	TRACE_MSG_INIT(128);
	sprintf(msg," error='%d' request='%d'/>\n", e->error_code,
		e->request_code);
	TRACE_MSG_SET(
		"XError", 0, msg, t, CurrentTime);
}

void send_anystr(char *type, char *str)
{
	unsigned long t;
	t = get_time();

	TRACE_MSG_INIT(256);
	sprintf(msg,">%s<%s/>\n", str, type);
	TRACE_MSG_SET(
		type, 0, msg, t, CurrentTime);	
}
