#ifndef XGET_H
#define XGET_H

#include <stdlib.h>
#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>

#include "atoms.h"

typedef struct
{
	unsigned long left;
	unsigned long right;
	unsigned long top;
	unsigned long bottom;
} strut_t;

typedef struct
{
	unsigned long left;
	unsigned long right;
	unsigned long top;
	unsigned long bottom;
	strut_t start;
	strut_t end;
} partial_strut_t;

typedef struct
{
	unsigned long orientation;
	unsigned long columns;
	unsigned long rows;
	unsigned long starting_corner;
} desktop_layout_t;

void *get_property(
	Display *disp, Window win, atom_t atom, unsigned int *size);

/* transform an atoms list into a string */
char *get_atoms_string(Display *disp, Atom *atom, unsigned int size);

/* check that the wm is "net/win" compliant. return the "check"
 * window or None */
Window get_net_check_window(Display *disp);
Window get_win_check_window(Display *disp);

/* return the atoms supported by the wm regarding "net/win" compliance,
 * rsize return the nbr of atoms */
Atom *get_net_supported_atoms(Display *disp, unsigned int *rsize);
Atom *get_win_supported_atoms(Display *disp, unsigned int *rsize);

/* retourne la liste des fenetres, size donne le nombre de fentres */
Window *get_client_list (Display *disp, unsigned int *size);

/* retourne la liste des fenetres dans l'ordre du "stacking", size est le
 * nombre de fentres */
Window *get_stacking_list(Display *disp, unsigned int *ret_size);

/* retourne la fentre active (avec le focus) */
Window get_active_window(Display *disp);

/* retourne le bureau courrant */
unsigned long get_current_desktop(Display * disp);

/* retourne le nombre de bureau */
long get_number_of_desktops(Display *disp);

/* retourne la position du coin suprieur droit de la partie visible
 * du bureau */ 
Bool get_desktop_viewport(
	Display *disp, unsigned long *x, unsigned long *y);

/* retourne la taille du bureau courrant */
Bool get_desktop_geometry(
	Display *disp, unsigned long *w, unsigned long *h);

/* les noms des desktops */
unsigned char **get_desktop_names(Display *disp, unsigned int *nbr);

/* arrangement des desktops */
Bool get_desktop_layout(Display *disp, desktop_layout_t *dl);

/* */
Bool get_showing_desktop(Display *disp, unsigned long *bool);

/* */
Bool get_workarea(
	Display *disp, int *x, int *y, unsigned int *w, unsigned int *h);

/* */
Bool isVirtualRoot(Display *disp, Window wind);
Bool get_set_virtual_roots(Display *disp);

/* retourn le bureau de la fentre wind */ 
unsigned long get_window_desktop(Display * disp, Window wind);

/* retourne l'etat de la fenetre wind et son layer (-1 si le layer n'est pas
 * trouv) */
void get_state_and_layer_window(
	Display * disp, Window wind, unsigned long *new_state, int *new_layer);

/*  retourne le layer de la fenetre wind (-1 si le layer n'est pas trouv) */
void get_layer_window(Display * disp, Window wind, int *new_layer);

/* */
Bool get_wm_state(Display *disp, Window wind, unsigned long *state);

/* retourne le net/win type de la fenetre wind */
Atom *get_net_window_type(Display *disp, Window wind, unsigned int *size);
Atom *get_win_window_type(Display *disp, Window wind, unsigned int *size);
/* les deux types dans une chaine (une fentre peut avoir plusieurs net type) */
char *get_window_type_str(Display *disp, Window wind);

/* */
char *get_client_machine(Display *disp, Window win);

/* retourne le pid qui control la fentre ou 0 s'il n'est pas trouv */
unsigned long get_window_pid(Display *disp, Window win);

/* */
strut_t *get_strut(Display *disp, Window wind);
partial_strut_t *get_partial_strut(Display *disp, Window wind);

/* */
strut_t *get_frame_extents(Display *disp, Window wind);

/* */
Bool get_icon_geometry(
	Display *disp, Window wind,
	int *x, int *y, unsigned int *w, unsigned int *h);

/* */
Bool get_handle_icons(Display *disp, Window wind);

/* */
Bool get_allowed_actions(Display *disp, Window wind, unsigned long *allowed);

/* */
Window get_transientFor(Display *disp, Window w);

/* retourne le nom de la fenetre */
char *get_window_title (Display *disp, Window win, int *is_net_name);

/* return true if to_found is "related" to start */
Bool get_is_in_tree(Display *disp, Window start, Window to_found);

/* return the top level parent of wind (i.e the wm frame of wind if any) */
Window get_toplevel_window(Display *disp, Window wind);

/* donne la gomtrie de la frame de la fentre wind, retourne 0 si
 * la fonction chouessss */
Bool get_frame_geometry(
	Display *disp, Window wind, int *fx, int *fy, unsigned int *fw,
	unsigned int *fh, Window *frame, strut_t **frame_extents);

Bool get_client_window(Display *disp, Window wind, Window *tl);

/* Xinerama stuff */
#ifdef HAVE_XINERAMA
#define FScreenHaveXinerama 1
#include <X11/extensions/Xinerama.h>
#else
#define FScreenHaveXinerama 0
typedef struct
{
	int   screen_number;
	short x_org;
	short y_org;
	short width;
	short height;
} XineramaScreenInfo;
#define XineramaQueryScreens(a, b) 0
#define XineramaQueryExtension(da, b, c) 0
#define XineramaIsActive(a) 0
#endif

XineramaScreenInfo *get_xinerama_info(Display *disp, int *count);

#endif
