/* -*-c-*- */
/* Copyright Olivier Chapuis (CNRS), Matthieu Langet (CNRS), 2005.
 *
 * <chapuis@lri.fr> <mathieu.langet@lri.fr>
 *
 * This software is a computer program whose purpose is to log the
 * window management activity into some files.
 *
 * This software is governed by the CeCILL  license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 *
 */


#include "config.h"

#include <string.h>
#include <time.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>

#include "include/wmstate.h"

#include "global.h"
#include "atoms.h"
#include "xget.h"
#include "xwindow.h"

#if IS_WMWATCH
#include "ipc/clientsocket_c.h"
#include "watch.h"
#elif IS_WMTRACE
#include "trace.h"
#endif

xwindow list_xwindow[1024];
int list_xwindow_size = 0;

int stacking_list[1024];
int stacking_list_size = 0;

unsigned long xinput_mask = StructureNotifyMask | PropertyChangeMask | KeyPressMask | KeyReleaseMask;

Bool have_wm_iconify = False;

extern Display *disp;
extern Window *virtualRoots;
extern unsigned int virtualRootsNbr;

Bool set_wm_info(int event_time, Bool ewmh_compliant, Bool win_compliant)
{
	Window wnet = None, wwin = None;
	Atom *net_atoms = NULL, *win_atoms = NULL;
	unsigned int net_size, win_size;
	char *wm_name = NULL;
	char *net_atoms_str = NULL;
	char *win_atoms_str = NULL;
	int dummy;

	if (ewmh_compliant && (wnet = get_net_check_window(disp)) == None)
	{
		return False;
	}
	if (win_compliant && (wwin = get_win_check_window(disp)) == None)
	{
		return False;
	}

	if (ewmh_compliant &&
	    (net_atoms = get_net_supported_atoms(disp, &net_size)) == NULL)
	{
		return False;
	}

	if (win_compliant &&
	    (win_atoms = get_win_supported_atoms(disp, &win_size)) == NULL)
	{
		if (net_atoms)
		{
			free(net_atoms);
		}
		return False;
	}

	if (wnet)
	{
		wm_name = get_window_title(disp, wnet, &dummy);
	}
	if (wm_name == NULL && wwin)
	{
		wm_name = get_window_title(disp, wwin, &dummy);
	}

	/* see if we have a way for detecting iconic state */
	if (net_atoms)
	{
		int i;
		for(i = 0; i < net_size; i++)
		{
			if (net_atoms[i] == XInternAtom(
				    disp, "_NET_WM_STATE_HIDDEN", True))
			{
				have_wm_iconify = True;
				break;
			}
		}
	}
	
	if (net_atoms)
	{
		net_atoms_str = get_atoms_string(disp, net_atoms, net_size);
		free(net_atoms);
	}
	if (win_atoms)
	{
		win_atoms_str = get_atoms_string(disp, win_atoms, win_size);
		free(win_atoms);
	}
	
	send_wm_info(event_time,
		(wm_name)? wm_name:"unknown",
		(net_atoms_str)? net_atoms_str:"no",
		(win_atoms_str)? win_atoms_str:"no");

	if (net_atoms_str)
	{
		free(net_atoms_str);
	}
	if (win_atoms_str)
	{
		free(win_atoms_str);
	}
	if (wm_name)
	{
		free(wm_name);
	}

	return True;
}

void set_stacking_list(int event_time)
{
	Window *list;
	unsigned int size;
	int i,sl_size = 0;
	Bool changed = False;

	list = get_stacking_list(disp, &size);

	if (size == 0 || list == NULL)
	{
		return;
	}

	for(i = 0; i < size; i++)
	{
		int id;
		if ((id = get_id_xwindow(list[i])) == -1)
		{
			continue;
		}
		if (sl_size > stacking_list_size ||
		    stacking_list[sl_size] != list_xwindow[id].uid_)
		{
			stacking_list[sl_size] = list_xwindow[id].uid_;
			changed = True;
		}
		sl_size++;
	}

	if (sl_size != stacking_list_size)
	{
		changed = True;
	}

	free(list);

	if (!changed)
	{
		return;
	}

	stacking_list_size = sl_size;
	
	send_stacking_list(event_time, stacking_list, stacking_list_size);
}

void set_active_xwindow(int event_time)
{
	static int active_window_uid = -1;
	static int active_window_id = -1;
	Window w;
	int id;

	w = get_active_window(disp);
	if ((id = get_id_xwindow(w)) == -1)
	{
		id = ROOT_ID;
	}
	if (list_xwindow[id].uid_ != active_window_uid)
	{
		if (active_window_id >= 0)
		{
			list_xwindow[active_window_id].active_ = 0;
		}
		list_xwindow[id].active_ = 1;
		active_window_uid = list_xwindow[id].uid_;
		active_window_id = id;
		send_simple_msg(ActiveWindow, id, event_time);
	}
}

void set_focusin_xwindow(int event_time, int id)
{
	if (list_xwindow[id].focus_ != 1)
	{
		list_xwindow[id].focus_ = 1;
		send_simple_msg(TakeFocus, id, event_time);
	}
}

void set_focusout_xwindow(int event_time, int id) {
	if (list_xwindow[id].focus_ != 0)
	{
		list_xwindow[id].focus_ = 0;
		send_simple_msg(ReleaseFocus, id, event_time);
	}
}

void set_current_desktop(int event_time)
{
	static unsigned long *current_desktop = NULL;
	unsigned long cd;

	cd = get_current_desktop(disp);

	if (current_desktop != NULL && *current_desktop == cd)
	{
		return;
	}
	if (current_desktop == NULL)
	{
		current_desktop = (unsigned long *)malloc(
			sizeof(unsigned long));
	}
	*current_desktop = cd;
	send_unsigned_msg(ChangeCurrentDesktop, ROOT_ID, event_time, cd);
}

void set_number_of_desktops(int event_time)
{
	static unsigned long *number_of_desktops = NULL;
	long nd;

	nd = get_number_of_desktops(disp);

	if (nd < 1)
	{
		return;
	}
	if (number_of_desktops != NULL &&  *number_of_desktops == nd)
	{
		return;
	}
	if (number_of_desktops == NULL)
	{
		number_of_desktops = (unsigned long *)malloc(
			sizeof(unsigned long));
	}
	*number_of_desktops = nd;
	send_unsigned_msg(NumberOfDesktops, ROOT_ID, event_time, nd); 
}

void set_desktop_viewport(int event_time)
{
	static unsigned long *desktop_viewport = NULL;
	unsigned long x,y;

	if (!get_desktop_viewport(disp, &x, &y))
	{
		return;
	}

	if (desktop_viewport != NULL && desktop_viewport[0] == x &&
	    desktop_viewport[1] == y)
	{
		return;
	}
	if (desktop_viewport == NULL)
	{
		desktop_viewport= (unsigned long *)malloc(
			2*sizeof(unsigned long));
	}

	desktop_viewport[0] = x;
	desktop_viewport[1] = y;
	
	send_position(DesktopViewport, ROOT_ID, event_time, (long)x, (long)y);
}

void set_desktop_geometry(int event_time)
{
	static unsigned long *desktop_geometry = NULL;
	unsigned long w = 0, h = 0;

	if (!get_desktop_geometry(disp, &w, &h))
	{
		return;
	}
	if (desktop_geometry != NULL && desktop_geometry[0] == w &&
	    desktop_geometry[1] == h)
	{
		return;
	}

	if (desktop_geometry == NULL)
	{
		desktop_geometry = (unsigned long *)malloc(
			2*sizeof(unsigned long));
	}

	desktop_geometry[0] = w;
	desktop_geometry[1] = h;

	send_size(DesktopGeometry, ROOT_ID, event_time, w, h);
}

#if IS_WMTRACE
static
void free_names_list(unsigned char **names, unsigned int nbr)
{
	int i;
	
	if (names == NULL)
	{
		return;
	}
	for(i = 0; i < nbr; i++)
	{
		if (names[i] != NULL)
		{
			free(names[i]);
		}
	}
	free(names);
	
}
#endif

void set_desktop_names(int event_time)
{
#if IS_WMTRACE
	static unsigned char **desktop_names = NULL;
	static unsigned int nbr_of_desktop_names = 0;
	unsigned char **names;
	unsigned int nbr;
	int i;
	
	names = get_desktop_names(disp, &nbr);

	if (names == NULL && desktop_names == NULL)
	{
		return;
	}

	if (names != NULL && desktop_names != NULL &&
	    nbr == nbr_of_desktop_names)
	{
		Bool changed = False;
		for(i = 0; i < nbr; i++)
		{
			if ((names[i] == NULL && desktop_names[i] != NULL) ||
			    (names[i] != NULL && desktop_names[i] == NULL) ||
			    (names[i] != NULL && desktop_names[i] != NULL &&
			     strcmp(names[i],desktop_names[i]) != 0))
			{
				changed = True;
				break;
			}
		}
		if (!changed)
		{
			free_names_list(names, nbr);
			return;
		}
	}

	if (desktop_names != NULL)
	{
		free_names_list(desktop_names, nbr_of_desktop_names);
	}
	desktop_names = names;
	nbr_of_desktop_names = nbr;
	if (desktop_names == NULL)
	{
		nbr_of_desktop_names = 0;
	}
	
	send_names_list(DesktopNames, ROOT_ID, event_time, names, nbr);
#endif
}

void set_desktop_layout(int event_time)
{
#if IS_WMTRACE
	static desktop_layout_t *desktop_layout = NULL;
	desktop_layout_t *dl;

	dl = (desktop_layout_t *)malloc(sizeof(desktop_layout_t));

	if (!get_desktop_layout(disp, dl))
	{
		free(dl);
		return;
	}
	if (desktop_layout != NULL &&
	    desktop_layout->orientation == dl->orientation &&
	    desktop_layout->columns == dl->columns &&
	    desktop_layout->rows == dl->rows &&
	    desktop_layout->starting_corner == dl->starting_corner)
	{
		free(dl);
		return;
	}

	if (desktop_layout)
	{
		free(desktop_layout);
	}
	desktop_layout = dl;
	send_desktop_layout(event_time, desktop_layout);
#endif
}

void set_showing_desktop(int event_time)
{
	static unsigned long *showing_desktop = NULL;
	unsigned long v;

	if (!get_showing_desktop(disp, &v))
	{
		return;
	}

	if (showing_desktop != NULL && *showing_desktop == v)
	{
		return;
	}

	if (showing_desktop == NULL)
	{
		showing_desktop = (unsigned long *)malloc(sizeof(unsigned long));
	}

	*showing_desktop = v;

	send_unsigned_msg(ShowingDesktop, 0, event_time, v);
}

void set_workarea(int event_time)
{
	static unsigned long *workarea = NULL;
	int x, y;
	unsigned int w, h;

	if (!get_workarea(disp, &x, &y, &w, &h))
	{
		return;
	}
	if (workarea != NULL && workarea[0] == x &&
	    workarea[1] == y && workarea[2] == h &&
	    workarea[3] == w)
	{
		return;
	}

	if (workarea == NULL)
	{
		workarea = (unsigned long *)malloc(4*sizeof(unsigned long));
	}

	workarea[0] = x;
	workarea[1] = y;
	workarea[2] = w;
	workarea[3] = h;

	send_geometry(DesktopWorkarea, ROOT_ID, event_time, x, y, w, h);
}

void set_virtual_roots(int event_time)
{
	int i;

	if (virtualRoots != NULL)
	{
		for (i = 0; i < virtualRootsNbr; i++)
		{
			XSelectInput(disp, virtualRoots[i], NoEventMask);
		}
	}

	get_set_virtual_roots(disp);

	if (virtualRoots != NULL)
	{
		for (i = 0; i < virtualRootsNbr; i++)
		{
			XSelectInput(disp, virtualRoots[i], xinput_mask);
		}
	}
}

void set_desktop_xwindow(int event_time, int id)
{
	unsigned long d;

	d = get_window_desktop(disp, list_xwindow[id].window_);
	if (d == list_xwindow[id].desktop_)
	{
		return;
	}
	list_xwindow[id].desktop_ = d;
	send_unsigned_msg(ChangeDesktopWindow, id, event_time, d);
}

void set_state_xwindow(int event_time, int id) {
	unsigned long new_state;
	int new_layer;
	int old_state = list_xwindow[id].state_;
	Bool shade_state_changed = False;

	get_state_and_layer_window(
		disp,list_xwindow[id].window_,&new_state, &new_layer);

	if (new_layer != -1 && new_layer != list_xwindow[id].layer_)
	{
		list_xwindow[id].layer_ = new_layer;
		send_unsigned_msg(
			LayerChange, id, event_time, (unsigned long)new_layer);
	}
	if (new_state != old_state)
	{
		
		list_xwindow[id].state_ = new_state;
		if ((old_state & WM_STATE_STICKY) !=
		    (new_state & WM_STATE_STICKY))
		{
			send_simple_msg(
				(new_state & WM_STATE_STICKY)?
				AlwaysVisibleWindow:NotAlwaysVisibleWindow,
				id, event_time);
		}
		if ((old_state & WM_STATE_MINIMIZED)!=
		    (new_state & WM_STATE_MINIMIZED ))
		{
			send_simple_msg(
				(new_state & WM_STATE_MINIMIZED)?
				MinimizeWindow:UnMinimizeWindow,
				id, event_time);
		}
		if ((old_state & WM_STATE_MAXIMIZED_VERT)!=
		    (new_state & WM_STATE_MAXIMIZED_VERT) || 
		    (old_state & WM_STATE_MAXIMIZED_HORIZ)!=
		    (new_state & WM_STATE_MAXIMIZED_HORIZ))
		{
			int max = (new_state &WM_STATE_MAXIMIZED_VERT)||
				(new_state & WM_STATE_MAXIMIZED_HORIZ);
			send_simple_msg(
				(max)? MaximizeWindow:UnMaximizeWindow,
				id, event_time);
		}
		if ((old_state & WM_STATE_HIDDEN) !=
		    (new_state & WM_STATE_HIDDEN))
		{
			send_simple_msg(
				(new_state & WM_STATE_HIDDEN)?
				HideWindow:UnHideWindow,id, event_time);
		}
		if ((old_state & WM_STATE_SHADED) !=
		    (new_state & WM_STATE_SHADED))
		{
			send_simple_msg(
				(new_state & WM_STATE_SHADED)?
				ShadeWindow:UnShadeWindow,id, event_time);
			shade_state_changed = True;
		}
		if ((old_state & WM_STATE_HID_WORKSPACE)!=
		    (new_state & WM_STATE_HID_WORKSPACE))
		{
			/* FIXME */
			send_unsigned_msg(
				ChangeDesktopWindow, id, event_time,
				list_xwindow[id].desktop_);
		}
		/* skip WM_STATE_HID_TRANSIENT  WM_STATE_FIXED_POSITION 
		 * WM_STATE_ARRANGE_IGNORE WM_STATE_SKIP_FOCUS
		 * WM_STATE_SKIP_WINLIST which are WIN specific any way */
		if ((old_state & WM_STATE_FULLSCREEN) !=
		    (new_state & WM_STATE_FULLSCREEN))
		{
			send_simple_msg(
				(new_state & WM_STATE_FULLSCREEN)?
				FullScreenWindow:UnFullScreenWindow,
				id, event_time);
		}
		if ((old_state & WM_STATE_SKIP_TASKBAR)!=
		    (new_state & WM_STATE_SKIP_TASKBAR))
		{
			send_simple_msg(
				(new_state & WM_STATE_SKIP_TASKBAR)?
				SkipTaskBarWindow:UnSkipTaskBarWindow,
				id, event_time);
		}
		/* skip WM_STATE_GROUP_TRANSIENT WM_STATE_FOCUS_ON_CLICK
		 * which are WIN specific any way */
		if ((old_state & WM_STATE_MODAL)!=
		    (new_state & WM_STATE_MODAL))
		{
			send_simple_msg(
			  (new_state & WM_STATE_MODAL)?
			  ModalWindow:UnModalWindow, id, event_time);
		}
		if ((old_state & WM_STATE_SKIP_PAGER)!=
		    (new_state & WM_STATE_SKIP_PAGER))
		{
			send_simple_msg(
				(new_state & WM_STATE_SKIP_PAGER)?
				SkipPagerWindow:UnSkipPagerWindow,
				id, event_time);
		}
	}
	/* FIXME: a voir avec mathieu (oli) */
	if (shade_state_changed)
	{
		set_geometry_xwindow(event_time,id);
	}
}


void set_wm_state_xwindow(int event_time, int id)
{
	unsigned long state;
	int old_state = list_xwindow[id].state_;

	if (have_wm_iconify)
	{
		return;
	}

	if (!get_wm_state(disp, list_xwindow[id].window_, &state))
	{
		return;
	}

	if (state == IconicState && !(old_state & WM_STATE_HIDDEN))
	{
		list_xwindow[id].state_ |= WM_STATE_HIDDEN;
		send_simple_msg(HideWindow, id, event_time);
	}
	else if (state == NormalState && (old_state & WM_STATE_HIDDEN))
	{
		list_xwindow[id].state_ &= ~WM_STATE_HIDDEN;
		send_simple_msg(UnHideWindow, id, event_time);
	}
}

void set_layer_xwindow(int event_time, int id)
{
	int new_layer;

	get_layer_window(disp,list_xwindow[id].window_,&new_layer);

	if (new_layer != -1 && new_layer != list_xwindow[id].layer_)
	{
		list_xwindow[id].layer_ = new_layer;
		send_unsigned_msg(
			LayerChange, id, event_time, (unsigned long)new_layer);
	}
}

void set_strut(int event_time, int id)
{
	strut_t *sstrut;
	strut_t *nstrut;

	sstrut = list_xwindow[id].strut_;
	nstrut = get_strut(disp, list_xwindow[id].window_);

	if (sstrut == NULL &&  nstrut != NULL)
	{
		list_xwindow[id].strut_ = nstrut;
		send_strut(WindowStrut, id, event_time, nstrut);
		return;
	}
	if (sstrut != NULL &&  nstrut == NULL)
	{
		free(sstrut);
		list_xwindow[id].strut_ = nstrut;
		send_strut(WindowStrut, id, event_time, nstrut);
		return;
	}
	if (sstrut == NULL ||  nstrut == NULL)
	{
		return;
	}

	if (sstrut->left != nstrut->left ||
	    sstrut->right != nstrut->right ||
	    sstrut->top != nstrut->top ||
	    sstrut->bottom != nstrut->bottom)
	{
		free(sstrut);
		list_xwindow[id].strut_ = nstrut;
		send_strut(WindowStrut, id, event_time, nstrut);
	}

}

void set_partial_strut(int event_time, int id)
{
	partial_strut_t *sstrut;
	partial_strut_t *nstrut;

	sstrut = list_xwindow[id].partial_strut_;
	nstrut = get_partial_strut(disp, list_xwindow[id].window_);

	if (sstrut == NULL &&  nstrut != NULL)
	{
		list_xwindow[id].partial_strut_ = nstrut;
		send_partial_strut(PartialWindowStrut, id, event_time, nstrut);
		return;
	}
	if (sstrut != NULL &&  nstrut == NULL)
	{
		free(sstrut);
		list_xwindow[id].partial_strut_ = nstrut;
		send_partial_strut(PartialWindowStrut, id, event_time, nstrut);
		return;
	}
	if (sstrut == NULL ||  nstrut == NULL)
	{
		return;
	}

	if (sstrut->left != nstrut->left ||
	    sstrut->right != nstrut->right ||
	    sstrut->top != nstrut->top ||
	    sstrut->bottom != nstrut->bottom ||
	    sstrut->start.left != nstrut->start.left ||
	    sstrut->start.right != nstrut->start.right ||
	    sstrut->start.top != nstrut->start.top ||
	    sstrut->start.bottom != nstrut->start.bottom ||
	    sstrut->end.left != nstrut->end.left ||
	    sstrut->end.right != nstrut->end.right ||
	    sstrut->end.top != nstrut->end.top ||
	    sstrut->end.bottom != nstrut->end.bottom)
	{
		free(sstrut);
		list_xwindow[id].partial_strut_ = nstrut;
		send_partial_strut(PartialWindowStrut, id, event_time, nstrut);
	}
}

void set_icon_geometry(int event_time, int id)
{
	int x,y;
	unsigned int w,h;

	get_icon_geometry(disp,list_xwindow[id].window_, &x, &y, &w, &h);

	if (x != list_xwindow[id].icon_x_ || y != list_xwindow[id].icon_y_ ||
	    w != list_xwindow[id].icon_w_ || h != list_xwindow[id].icon_h_)
	{
		list_xwindow[id].icon_x_ = x;
		list_xwindow[id].icon_y_ = y;
		list_xwindow[id].icon_w_ = w;
		list_xwindow[id].icon_h_ = h;
		send_geometry(IconGeometry, id, event_time, x, y, w, h);
	}
}

void set_handle_icons(int event_time, int id)
{
	Bool h = get_handle_icons(disp, list_xwindow[id].window_);

	if (h != list_xwindow[id].handle_icons_)
	{
		list_xwindow[id].handle_icons_ = h;
		send_unsigned_msg(HandleIcons, id, event_time, h);
	}

}

void set_allowed_actions(int event_time, int id)
{
	long allowed;

	get_allowed_actions(disp, list_xwindow[id].window_, &allowed);

	if (allowed != list_xwindow[id].allowed_actions_)
	{
		list_xwindow[id].allowed_actions_ = allowed;
		send_long_msg(AllowedActions, id, event_time, allowed);
	}

}

void set_name_xwindow(int event_time, int id)
{
	char *name;
	int is_net_name;

	name = get_window_title(disp, list_xwindow[id].window_, &is_net_name);

	/* name cannot be null */
	if (list_xwindow[id].is_net_name_ == is_net_name &&
	    strcmp(name,list_xwindow[id].name_) == 0)
	{
		free(name);
		return;
	}

	free(list_xwindow[id].name_);
	list_xwindow[id].is_net_name_ = is_net_name;
	list_xwindow[id].name_ = name;

	send_name(ChangeNameWindow, id, event_time, name, is_net_name);
}

void set_selection_notify(int event_time, Window wind, atom_t atom)
{
	char *val,*rep;
	static char *repold = NULL;
	unsigned int size;
	Window selection_owner;
	int selection_owner_id;

	if ((val = get_property(disp, wind, atom, &size)) == NULL)
	{
		return;
	}

	rep = malloc(sizeof(char)*size+1);
	memcpy(rep,val,size);
	rep[size] = '\0';
	free(val);

	if (rep != NULL && (repold == NULL || strcmp(rep,repold) != 0))
	{
		if (repold) free(repold);
		repold = rep;

		selection_owner = XGetSelectionOwner(disp, XA_PRIMARY);
		selection_owner_id = -1;
		if (selection_owner != None)
		{
			selection_owner_id =
				get_id_xwindow_in_tree(selection_owner);
		}
		if (selection_owner_id == -1)
		{
			selection_owner_id = ROOT_ID;
		}
		send_selection_notify(selection_owner_id, event_time, rep);
        }
	else if (rep != NULL)
	{
		free(rep);
	}

}

void set_selection_request(int event_time, Window requestor, Window owner)
{
	int requestor_id, owner_id;

	requestor_id = get_id_xwindow_in_tree(requestor);
	if (requestor_id == -1)
	{
		requestor_id = ROOT_ID;
	}
	owner_id = get_id_xwindow_in_tree(owner);
	if (owner_id == -1)
	{
		owner_id = ROOT_ID;
	}
	send_selection_request(event_time, owner_id, requestor_id);
}

void set_button_event(
	int event_time, Window wind, int type, Window child, int x, int y,
	int button, int state)
{
	static button_event *prev_be = NULL;
	int id = -1;
	Window client;

	/* filter */
	if (!isVirtualRoot(disp, wind) && child != None &&
	    get_client_window(disp, wind, &client) && client != wind)
	{
		return;
	}

	if (child != None)
	{
		id = get_id_xwindow_in_tree(child);
	}
	else if (isVirtualRoot(disp, wind))
	{
		id = ROOT_ID;
	}
	else
	{
		id = get_id_xwindow_in_tree(wind);
	}
	if (id == -1)
	{
		return;
	}
	if (prev_be == NULL)
	{
		prev_be = (button_event *)malloc(sizeof(button_event));
	}
	else if (prev_be->type == type &&
		 prev_be->uid == list_xwindow[id].uid_ &&
		 prev_be->event_time == event_time &&
		 prev_be->x == x &&
		 prev_be->y == y &&
		 prev_be->button == button &&
		 prev_be->state == state)
	{
		return;
	}
	prev_be->type = type;
	prev_be->uid = list_xwindow[id].uid_;
	prev_be->event_time = event_time;
	prev_be->x = x;
	prev_be->y = y;
	prev_be->button = button;
	prev_be->state = state;
	send_button_msg(event_time, prev_be);
}

/* FIXME: get the button and the state */
void set_fake_button_event(int event_time, Window win)
{
	Window retour_racine, retour_fille;
	int retour_racine_x, retour_racine_y;
	int retour_fen_x, retour_fen_y;
	unsigned int retour_masque;
	button_event be;

	if (!XQueryPointer(
		    disp, win, &retour_racine, &retour_fille,
		    &retour_racine_x, &retour_racine_y,
		    &retour_fen_x, &retour_fen_y, &retour_masque))
	{
		return;
	}
	be.type = wmButtonMotion;
	be.uid = ROOT_ID;
	be.event_time = event_time;
	be.x = retour_racine_x;
	be.y = retour_racine_y;
	be.button = 0;
	be.state = 0;
	send_button_msg(event_time, &be);
}


void set_key_event(int event_time, int type, Window wind, int key, int state)
{
	static key_event *prev_ke = NULL;
	int id = -1;

	if (isVirtualRoot(disp, wind))
	{ 
		id = ROOT_ID;
	}
	else
	{
		id = get_id_xwindow_in_tree(wind);
	}
	if (id == -1)
	{
		return;
	}
	if (prev_ke == NULL)
	{
		prev_ke = (key_event *)malloc(sizeof(key_event));
	}
	else if (prev_ke->type == type &&
		 prev_ke->uid == list_xwindow[id].uid_ &&
		 prev_ke->event_time == event_time &&
		 prev_ke->key == key &&
		 prev_ke->state == state)
	{
		return;
	}
	prev_ke->type = type;
	prev_ke->uid = list_xwindow[id].uid_;
	prev_ke->event_time = event_time;
	prev_ke->key = key;
	prev_ke->state = state;
	send_key_msg(type, id, event_time, key, state);
}

void set_geometry_xwindow(int event_time, int id)
{
	int x, y;
	unsigned int w, h;
	Window frame;
	strut_t *fe = NULL;
	xwindow *xwin = &list_xwindow[id];
	
	if (!get_frame_geometry(
		    disp, xwin->window_, &x, &y, &w, &h, &frame, &fe))
	{
		/* no more window */
		return;
	}
	/* FIXME: handle frame change */
	xwin->frame_ = frame;

	if ((xwin->x_ != x || xwin->y_  != y) &&
	    (xwin->w_ != w || xwin->h_  != h))
	{
		send_geometry(
			MoveResizeWindow, id, event_time, x, y, w, h);
	}
	else if (xwin->x_ != x || xwin->y_  != y)
	{
		send_position(MoveWindow, id, event_time, x, y);
	}
	else if (xwin->w_ != w || xwin->h_  != h)
	{
		send_size(
			ResizeWindow, id, event_time, (unsigned long)w,
			(unsigned long)h);
	}
	if (fe != NULL &&
	    (xwin->frame_extents_ == NULL ||
	     xwin->frame_extents_->right != fe->right ||
	     xwin->frame_extents_->left != fe->left ||
	     xwin->frame_extents_->top != fe->top ||
	     xwin->frame_extents_->bottom != fe->bottom))
	{
		send_strut(FrameExtents, id, event_time, fe);
		if (xwin->frame_extents_ == NULL)
		{
			xwin->frame_extents_ =
				(strut_t *)malloc(sizeof(strut_t));
		}
		xwin->frame_extents_->right = fe->right;
		xwin->frame_extents_->left = fe->left;
		xwin->frame_extents_->top = fe->top;
		xwin->frame_extents_->bottom = fe->bottom;
	}
	if (fe)
	{
		free(fe);
	}
	xwin->h_ = h;
	xwin->w_ = w;
	xwin->x_ = x;
	xwin->y_ = y;
}


void set_xinerama_info(int event_time)
{
	XineramaScreenInfo *info;
	int count;

	if ((info = get_xinerama_info(disp, &count)) == NULL || count == 0)
	{
		return;
	}

	send_xinerama_info(event_time, info, count);

	XFree(info);
}

/*
 *
 * list_xwindow management
 *
 */

int create_xwindow(
	Window wind, Window frame, Window transientFor, long desktop,
	long pid, int x, int y, int w, int h, strut_t *frame_extents,
	char* machine, XWMHints *wmhints,
	XClassHint classhints, char* name, int is_net_name, char* type,
	unsigned long state, int layer, strut_t *strut,
	partial_strut_t *partial_strut,  int ix, int iy, unsigned int iw,
	unsigned int ih, Bool handle_icons, long allowed_actions, int management,
	int event_time)
{
  static unsigned long uid = 0;
  xwindow xwin;

  xwin.window_ = wind;
  xwin.frame_ = frame;
  xwin.transientFor_ = transientFor;
  xwin.uid_ = uid++;
  xwin.desktop_ = desktop;
  xwin.pid_ = pid;
  xwin.x_ = x;
  xwin.y_ = y;
  xwin.w_ = w;
  xwin.h_ = h;
  xwin.name_ = name;
  xwin.is_net_name_ = is_net_name;
  xwin.machine_ = machine;
  xwin.wmhints_ = wmhints;
  xwin.classhints_ = classhints;
  xwin.focus_ = 0;
  xwin.active_ = 0;
  xwin.type_ = type;
  xwin.state_ = state;
  xwin.layer_ = layer;
  xwin.strut_ = strut;
  xwin.partial_strut_ = partial_strut;
  xwin.frame_extents_ = frame_extents;
  xwin.icon_x_ = ix;
  xwin.icon_y_ = iy;
  xwin.icon_w_ = iw;
  xwin.icon_h_ = ih;
  xwin.handle_icons_ = handle_icons;
  xwin.allowed_actions_ = allowed_actions;
  xwin.management_ = management; 
  list_xwindow[list_xwindow_size] = xwin;

  send_create_window(list_xwindow_size, event_time);
  list_xwindow_size++;

  return (list_xwindow_size - 1);
}

void x_select_input(Display* disp, Window wind)
{
	XSelectInput(disp, wind, xinput_mask);

	return;
}

static
void add_or_xwindow(int event_time, Window wind, int management)
{
	int id;
	/*int x, y;
	unsigned int width, height, bw, depth;
	Window junkroot; */
	XClassHint classhints;
	XWindowAttributes winattrs = {0};
	XEvent e;

	/*fprintf(stderr,"XGetWindowAttributes\n");*/
	XSync(disp,0);
	/*fprintf(stderr,"XSynced\n"); */
	while(XCheckTypedWindowEvent(
		      disp, wind, ConfigureNotify, &e)) {}
	/*fprintf(stderr,"Flushed\n");*/
	if (!XGetWindowAttributes(disp, wind, &winattrs))
	{
		/*fprintf(stderr,"FAIL\n");*/
		return;
	}
	/*fprintf(stderr,"GET IT\n");*/

	classhints.res_class = NULL;
	classhints.res_name = NULL;

	id = create_xwindow(
	  wind,
	  wind, /*frame*/
	  None, /*transientFor*/
	  0, /* desktop */
	  0, /* pid */
	  /* x, y, width, height, */
	  winattrs.x, winattrs.y, winattrs.width, winattrs.height,
	  NULL, /* frame_extents */
	  NULL, /*client_machine*/
	  NULL /*wmhints*/,
	  classhints,
	  NULL, /*title*/
	  0, /*is_net_title*/
	  NULL, /*type*/
	  0, /*state*/
	  0, /*layer*/
	  NULL, /*strut*/
	  NULL, /*partial_strut*/
	  0,0,0,0, /*icon geometry*/
	  0, /*handle_icon*/
	  -1, /*allowed_actions*/
	  management,
	  event_time);
}

static
void add_xwindow(int event_time, Window wind, int management)
{
  char *title;
  int is_net_title;
  char *client_machine;
  unsigned long pid;
  unsigned long desktop;
  int x, y;
  unsigned int wwidth, wheight;
  Window frame;
  char* type;
  unsigned long state, wmstate;
  int layer;
  int id;
  XWMHints *wmhints = NULL;
  XClassHint classhints;
  Window transientFor = None;
  strut_t *strut;
  partial_strut_t *partial_strut;
  strut_t *frame_extents;
  int ix, iy;
  unsigned int iw, ih;
  Bool handle_icons;
  long allowed_actions;

  x_select_input(disp,wind);

  title = get_window_title(disp, wind, &is_net_title);

  /* desktop ID */
  desktop = get_window_desktop(disp,wind);

  /* client machine */
  client_machine = get_client_machine(disp, wind);

  /* pid */
  pid = get_window_pid(disp, wind);

  /* geometry */
  if (!get_frame_geometry(
	      disp, wind, &x, &y, &wwidth, &wheight, &frame, &frame_extents))
  {
	  x = y = 0;
	  wwidth = wheight = 10;
  }
  if (frame != None && wind != frame)
  {
	  XSelectInput(disp, frame, StructureNotifyMask);
  }

  wmhints = XGetWMHints(disp, wind);
  classhints.res_class = NULL;
  classhints.res_name = NULL;
  XGetClassHint(disp, wind, &classhints);

  transientFor = get_transientFor(disp, wind);

  /* type */
  type = get_window_type_str(disp, wind);

  /* state */
  get_state_and_layer_window(disp,wind,&state,&layer);

  if (!have_wm_iconify && get_wm_state(disp,wind,&wmstate) &&
      wmstate == IconicState)
  {
	  state |= WM_STATE_HIDDEN;
  }

  if (layer == -1)
  {
	  get_layer_window(disp,wind,&layer);
  }

  /* strut */
  strut = get_strut(disp, wind);
  partial_strut = get_partial_strut(disp, wind);

  /* icon geometry */
  get_icon_geometry(disp, wind, &ix, &iy, &iw, &ih);

  handle_icons = get_handle_icons(disp, wind);

  get_allowed_actions(disp, wind, &allowed_actions);

  id = create_xwindow(
	  wind,frame,transientFor,desktop,pid,x,y,wwidth,wheight,frame_extents,
	  client_machine, wmhints,classhints,title,is_net_title, type,state,
	  layer, strut, partial_strut, ix,iy,iw,ih, handle_icons,
	  allowed_actions, management, event_time);

}

static
void remove_xwindow(int uid, int event_time)
{
	int i = 1;

	while(i < list_xwindow_size && uid != list_xwindow[i].uid_)
	{
		i++;
	}

	if (i == list_xwindow_size)
	{
		return;
	}

	list_xwindow_size--;
	send_simple_msg(DestroyWindow, i, event_time);

	free(list_xwindow[i].name_);
	free(list_xwindow[i].machine_);
	free(list_xwindow[i].type_);
	if (list_xwindow[i].classhints_.res_class)
		XFree ((char *)list_xwindow[i].classhints_.res_class);
	if (list_xwindow[i].classhints_.res_name)
		XFree ((char *)list_xwindow[i].classhints_.res_name);
	if (list_xwindow[i].wmhints_)
		XFree(list_xwindow[i].wmhints_);
	if (list_xwindow[i].strut_)
		free(list_xwindow[i].strut_);
	if (list_xwindow[i].partial_strut_)
		free(list_xwindow[i].partial_strut_);
	if (list_xwindow[i].frame_extents_)
		free(list_xwindow[i].frame_extents_);
	if (i != list_xwindow_size)
	{
		memcpy(&list_xwindow[i], &list_xwindow[list_xwindow_size],
		       sizeof(list_xwindow[list_xwindow_size]));
	}
}

void check_xwindow(int event_time)
{
	Window *client_list;
	unsigned int client_list_size;
	int i,j;
	Bool found;
	int remove_uid[1024];
	int remove_uid_size = 0;

	if ((client_list = get_client_list(disp, &client_list_size)) == NULL)
	{
		client_list_size = 0;
	}

	for (i = 0; i < client_list_size; i++)
	{
		if (get_id_xwindow(client_list[i]) == -1)
		{
			add_xwindow(event_time, client_list[i], MM_MANAGED);
		}
	}

	/* never remove the root window */
	for (j = 1; j < list_xwindow_size; j++)
	{
		if (list_xwindow[j].management_ != MM_MANAGED)
		{
			continue;
		}
		i = 0;
		found = False;
		while(i < client_list_size && !found)
		{
			if (list_xwindow[j].window_ == client_list[i])
			{
				found = True;
			}
			i++;
		}
		if (!found)
		{
			remove_uid[remove_uid_size++] = list_xwindow[j].uid_;
		}
	}
	
	for (i = 0; i < remove_uid_size; i++)
	{
		remove_xwindow(remove_uid[i], event_time);
	}

	if (client_list)
	{
		free(client_list);
	}
}

void checkadd_unmanaged_xwindow(int event_time, Window wind, Bool or)
{
	int i;
	Window client;

	if (get_id_xwindow(wind) != -1)
	{
		return;
	}

	if (or)
	{
		for(i = 0; i < list_xwindow_size; i++)
		{
			/* icewm and kwin have or window frame! */
			if (list_xwindow[i].frame_ == wind)
			{
				return;
			}
		}
		/* check if this or window is a frame of a window */
		if (get_client_window(disp, wind, &client))
		{
			add_xwindow(event_time, client, MM_CACHED);
		}
		else
		{
			add_or_xwindow(event_time, wind, MM_OVERRIDE_REDIRECT);
		}
	}
}

void check_cached_xwindow(int event_time, Window Root)
{
	Window junkroot;
	Window parent;
	Window client;
	Window *filles = NULL;
	int nfilles,i;

	if (!XQueryTree(disp, Root, &junkroot, &parent, &filles, &nfilles))
	{
		if (filles)
		{
			XFree(filles);
		}
	}
	for(i = 0; i < nfilles; i++)
	{
		if (get_client_window(disp, filles[i], &client) &&
			get_id_xwindow(client) == -1)
		{
			add_xwindow(event_time, client, MM_CACHED);
		}
	}
}

void checkremove_unmanaged_xwindow(int event_time, Window wind)
{
	int id;

	if ((id = get_id_xwindow(wind)) == -1 ||
	    list_xwindow[id].management_ == MM_ROOTWIN ||
	    list_xwindow[id].management_ == MM_MANAGED)
	{
		return;
	}
	remove_xwindow(list_xwindow[id].uid_, event_time);
}

int get_id_xwindow(Window wind)
{
  int i = 0;

  if (isVirtualRoot(disp, wind))
  {
	  return ROOT_ID;
  }

  while (i < list_xwindow_size && list_xwindow[i].window_ != wind) {
    i++;
  }

  if (i < list_xwindow_size)
  {
	  return i;
  }
  return -1;
}

int get_id_xwindow_by_frame(Window wind)
{
  int i = 0;

  if (isVirtualRoot(disp, wind))
  {
	  return ROOT_ID;
  }

  while (i < list_xwindow_size && list_xwindow[i].frame_ != wind) {
    i++;
  }

  if (i < list_xwindow_size)
  {
	  return i;
  }
  return -1;
}

int get_id_xwindow_in_tree(Window wind)
{
  int i = 0;

  if ((i = get_id_xwindow(wind)) != -1)
  {
	  return i;
  }

  i = 0;
  while (i < list_xwindow_size)
  {
	  if (list_xwindow[i].frame_ == wind)
	  {
		  return i;
	  }
	  if (get_is_in_tree(disp, wind, list_xwindow[i].window_))
	  {
		  return i;
	  }
	  i++;
  }

  {
	  Window f1 = get_toplevel_window(disp, wind);
	  i = 0;
	  while (i < list_xwindow_size)
	  {
		  Window f2 = get_toplevel_window(disp, list_xwindow[i].window_);
		  if (f1 == f2)
		  {
			  return i;
		  }
		  i++;
	  }
  }

  return -1;
}


