#ifndef XWINDOW_H
#define XWINDOW_H

#include <X11/Xutil.h>

#include "atoms.h"

typedef struct
{
	char* name_;          /* nom de la fenetre */
	int is_net_name_;     /* 1 ou 2 si le nom provient d'une net hints
			       * (et est donc en utf8), 2 si c'est un "visible"
			       * name */
	char* machine_;       /* nom de la machine */
	XWMHints *wmhints_;
	XClassHint classhints_; /* ResourceName and ClassName */
	Window window_;       /* X id de la fenetre */
	Window frame_;        /* X id de la frame de la fenetre */
	Window transientFor_; /* si != None la fentre est transitoire pour cette
			       * fentre */
	unsigned long uid_;   /* id unique */
	int x_;               /* position en X de la fenetre */
	int y_;               /* position en Y de la fenetre */
	int h_;               /* hauteur de la fenetre */
	int w_;               /* largeur de la fenetre */
	long desktop_;        /* bureau ou se trouve la fenetre */
	long pid_;            /* processus execute dans cette fenetre
			       * (0 si c'est un terminal) */
	int focus_;           /* si la fenetre a le focus = 1 sinon = 0 */
	int active_;          /* si la fenetre est active = 1 */
	unsigned long state_; /* l'etat de la fenetre (reduite, maximize,
			       * fullscreen, shade, ...) */
	char* type_;          /* le type de la fenetre */
	int layer_;           /* le layer 4: normal */
	strut_t *strut_;      /* voir le wm-spec */
	partial_strut_t *partial_strut_;
	strut_t *frame_extents_;/* description de la decoration de la fentre */
	int icon_x_;          /* icon geometry */
	int icon_y_;
	unsigned int icon_h_;
	unsigned int icon_w_;
	Bool handle_icons_;   /* l'application conteint une taskbar */
	long allowed_actions_;
#define MM_MANAGED 0
#define MM_OVERRIDE_REDIRECT 1
#define MM_CACHED 2
#define MM_ROOTWIN 3
	int management_;      /* MM_MANAGED: in client list
			       * MM_OVERRIDE_REDIRECT: not in client list but or
			       * MM_CACHED: not in client list and not or!!
			       * MM_ROOTWIN: the root window
			       */
} xwindow;


typedef struct
{
	int type;
	long uid;
	int event_time;
	int x;
	int y;
	int button;
	long state;
} button_event;

typedef struct
{
	int type;
	long uid;
	int event_time;
	int key;
	int button;
	int state;
} key_event;

/* */
Bool set_wm_info(int event_time, Bool ewmh_compliant, Bool win_compliant);

/* */
void set_stacking_list(int event_time);

/* met  jour l'attribut active_(= 1) de la xwindow identifi par id */
void set_active_xwindow(int event_time);
/* met  jour l'attribut focus_(= 0) de la xwindow identifi par id */
void set_focusout_xwindow(int event_time, int id);
/* met  jour l'attribut focus_(= 1) de la xwindow identifi par id */
void set_focusin_xwindow(int event_time, int id);

/* */
void set_current_desktop(int event_time);

/* */
void set_number_of_desktops(int event_time);

/* */
void set_desktop_viewport(int event_time);

/* */
void set_desktop_geometry(int event_time);

/* */
void set_desktop_names(int event_time);

/* */
void set_desktop_layout(int event_time);

/* */
void set_showing_desktop(int event_time);

/* */
void set_workarea(int event_time);

/* met  jour l'attribut desktop_ de la xwindow identifi par id */
void set_desktop_xwindow(int event_time, int id);

/* */
void set_virtual_roots(int event_time);

/* met  jour l'attribut state_ de la xwindow identifi par id */
void set_state_xwindow(int event_time, int id);

/* */
void set_wm_state_xwindow(int event_time, int id);

/* met  jour l'attribut layer_ de la xwindow identifi par id */
void set_layer_xwindow(int event_time, int id);

/* */
void set_strut(int event_time, int id);

/* met  jour le titre de la fentre */
void set_name_xwindow(int event_time, int id);
void set_partial_strut(int event_time, int id);

/* */
void set_icon_geometry(int event_time, int id);

/* */
void set_handle_icons(int event_time, int id);

/* */
void set_allowed_actions(int event_time, int id);

/* */
void set_selection_notify(int event_time, Window wind, atom_t atom);

/* */
void set_selection_request(int event_time, Window requestor, Window owner);

/* */
void set_button_event(
	int event_time, Window wind, int type, Window child, int x, int y,
	int button, int state);
void set_fake_button_event(int event_time, Window win);

/* */
void set_key_event(int event_time, int type, Window wind, int key, int state);

/* met  jour l'attribut h_, w_ x_ et y_ de la xwindow identifi par id */
void set_geometry_xwindow(int event_time, int id);

/* */
void set_xinerama_info(int event_time);

/* cre une xwindow */
int create_xwindow(
	Window wind, Window frame, Window transientFor, long desktop,
	long pid, int x, int y, int w, int h, strut_t *frame_extents,
	char* machine, XWMHints *wmhints,
	XClassHint classhints, char* name, int is_net_name, char* type,
	unsigned long state, int layer, strut_t *strut,
	partial_strut_t *partial_strut,  int ix, int iy, unsigned int iw,
	unsigned int ih, Bool handle_icons, long allowed_actions, int management,
	int event_time);

/* */
void x_select_input(Display* disp, Window wind);

/* ajoute/supprime des xwindow a la liste en fonction de la client list */
void check_xwindow(int event_time);
void checkadd_unmanaged_xwindow(int event_time, Window wind, Bool or);
void check_cached_xwindow(int event_time, Window Root);
void checkremove_unmanaged_xwindow(int event_time, Window wind);

/* retourne l'identifiant xwindow de la fenetre wind si elle est prsente dans
 * la liste, et -1 sinon */
int get_id_xwindow(Window wind);
int get_id_xwindow_by_frame(Window wind);

/* */
int get_id_xwindow_in_tree(Window wind);

/* */
Window get_new_xwindow(Display* disp);

#endif
